import tensorflow.keras.backend as K
from tensorflow.keras.layers import Dense, Flatten, Conv2D, MaxPooling2D, Input, Activation, \
    Lambda, UpSampling2D, \
                                    Conv2DTranspose, Reshape, BatchNormalization
from tensorflow.keras.models import Model
from tensorflow.keras.losses import mse, binary_crossentropy

from miso.layers import cyclic
from miso.archive.datasource import DataSource
import numpy as np


def sampling(args):
    """Reparameterization trick by sampling fr an isotropic unit Gaussian.
    # Arguments
        args (tensor): mean and log of variance of Q(z|X)
    # Returns
        z (tensor): sampled latent vector
    """
    z_mean, z_log_var = args
    batch = K.shape(z_mean)[0]
    dim = K.int_shape(z_mean)[1]
    # by default, random_normal has mean=0 and std=1.0
    epsilon = K.random_normal(shape=(batch, dim))
    return z_mean + K.exp(0.5 * z_log_var) * epsilon


def sampling2(args):
    z_mean, z_log_sigma, n_size = args
    epsilon = K.random_normal(shape=(n_size,), mean=0, stddev=1)
    return z_mean + K.exp(z_log_sigma/2) * epsilon


def vae_model(input_shape):
    # network parameters
    # input_shape = (image_size, image_size, 1)
    # batch_size = 128
    blocks = 2
    kernel_size = 5
    filters = 16
    latent_dim = 2
    # epochs = 30

    # VAE model = encoder + decoder
    # build encoder model
    inputs = Input(shape=input_shape, name='encoder_input')
    x = inputs
    for i in range(blocks):
        filters *= 2
        x = Conv2D(filters=filters,
                   kernel_size=kernel_size,
                   activation='relu',
                   strides=2,
                   padding='same')(x)

    # shape info needed to build decoder model
    shape = K.int_shape(x)

    # generate latent vector Q(z|X)
    x = Flatten()(x)
    x = Dense(16, activation='relu')(x)
    z_mean = Dense(latent_dim, name='z_mean')(x)
    z_log_var = Dense(latent_dim, name='z_log_var')(x)

    # use reparameterization trick to push the sampling out as input
    # note that "output_shape" isn't necessary with the TensorFlow backend
    z = Lambda(sampling, output_shape=(latent_dim,), name='z')([z_mean, z_log_var])

    # instantiate encoder model
    encoder = Model(inputs, [z_mean, z_log_var, z], name='encoder')
    encoder.summary()
    # plot_model(encoder, to_file='vae_cnn_encoder.png', show_shapes=True)

    # build decoder model
    latent_inputs = Input(shape=(latent_dim,), name='z_sampling')
    x = Dense(shape[1] * shape[2] * shape[3], activation='relu')(latent_inputs)
    x = Reshape((shape[1], shape[2], shape[3]))(x)

    for i in range(blocks):
        x = Conv2DTranspose(filters=filters,
                            kernel_size=kernel_size,
                            activation='relu',
                            strides=2,
                            padding='same')(x)
        filters //= 2

    outputs = Conv2DTranspose(filters=1,
                              kernel_size=kernel_size,
                              activation='sigmoid',
                              padding='same',
                              name='decoder_output')(x)

    # instantiate decoder model
    decoder = Model(latent_inputs, outputs, name='decoder')
    decoder.summary()
    # plot_model(decoder, to_file='vae_cnn_decoder.png', show_shapes=True)

    # instantiate VAE model
    outputs = decoder(encoder(inputs)[2])
    vae = Model(inputs, outputs, name='vae')

    return vae, z_mean, z_log_var





def build_conv_vae(input_shape, bottleneck_size, samp):
    # ENCODER
    input = Input(shape=(input_shape[0], input_shape[1], input_shape[2]))
    x = Conv2D(32, (3, 3), activation='relu', padding='same')(input)
    x = BatchNormalization()(x)
    x = MaxPooling2D((2, 2), padding='same')(x)
    x = Conv2D(64, (3, 3), activation='relu', padding='same')(x)
    x = BatchNormalization()(x)
    x = MaxPooling2D((2, 2), padding='same')(x)
    x = Conv2D(128, (3, 3), activation='relu', padding='same')(x)
    x = BatchNormalization()(x)
    x = MaxPooling2D((2, 2), padding='same')(x)
    x = Conv2D(256, (3, 3), activation='relu', padding='same')(x)
    x = BatchNormalization()(x)
    x = MaxPooling2D((2, 2), padding='same')(x)

    # Latent Variable Calculation
    shape = K.int_shape(x)
    flatten_1 = Flatten()(x)
    dense_1 = Dense(bottleneck_size, name='z_mean')(flatten_1)
    z_mean = BatchNormalization()(dense_1)
    print(dense_1)
    print(z_mean)
    flatten_2 = Flatten()(x)
    dense_2 = Dense(bottleneck_size, name='z_log_sigma')(flatten_2)
    z_log_sigma = BatchNormalization()(dense_2)
    print(dense_2)
    print(z_log_sigma)
    z = Lambda(samp)([z_mean, z_log_sigma])
    encoder = Model(input, [z_mean, z_log_sigma, z], name='encoder')

    # DECODER
    latent_input = Input(shape=(bottleneck_size,), name='decoder_input')
    x = Dense(shape[1] * shape[2] * shape[3])(latent_input)
    x = Reshape((shape[1], shape[2], shape[3]))(x)
    x = UpSampling2D((2, 2))(x)
    # x = Cropping2D([[0, 0], [0, 1]])(x)
    x = Conv2DTranspose(256, (3, 3), activation='relu', padding='same')(x)
    x = BatchNormalization()(x)
    x = UpSampling2D((2, 2))(x)
    # x = Cropping2D([[0, 1], [0, 1]])(x)
    x = Conv2DTranspose(128, (3, 3), activation='relu', padding='same')(x)
    x = BatchNormalization()(x)
    x = UpSampling2D((2, 2))(x)
    # x = Cropping2D([[0, 1], [0, 1]])(x)
    x = Conv2DTranspose(64, (3, 3), activation='relu', padding='same')(x)
    x = BatchNormalization()(x)
    x = UpSampling2D((2, 2))(x)
    x = Conv2DTranspose(32, (3, 3), activation='relu', padding='same')(x)
    x = BatchNormalization()(x)
    output = Conv2DTranspose(1, (3, 3), activation='tanh', padding='same')(x)
    decoder = Model(latent_input, output, name='decoder')

    print(encoder.outputs)
    print(input)
    print(decoder(encoder.outputs[2]))
    output_2 = decoder(encoder.outputs[2])
    vae = Model(input, output_2, name='vae')
    return vae, encoder, decoder, z_mean, z_log_sigma


# def vae_loss(input_img, output, vae=None):
    # Compute error in reconstruction
    # reconstruction_loss = mse(K.flatten(input_img), K.flatten(output))
    #
    # # Compute the KL Divergence regularization term
    # kl_loss = - 0.5 * K.sum(1 + z_log_sigma - K.square(z_mean) - K.exp(z_log_sigma), axis=-1)
    #
    # # Return the average loss over all images in batch
    # total_loss = (reconstruction_loss + 0.0001 * kl_loss)





    # return total_loss


if __name__ == '__main__':

    input_shape = (64, 64, 1)
    bottleneck_size = 128
    image_size = 64
    # vae, z_mean, z_log_var = vae_model((64,64,1))

    vae, encoder, decoder, z_mean, z_log_sigma = \
        build_conv_vae(input_shape, bottleneck_size, sampling)

    # VAE loss = mse_loss or xent_loss + kl_loss
    # if False:
    #     reconstruction_loss = mse(K.flatten(inputs), K.flatten(outputs))
    # else:
    # reconstruction_loss = binary_crossentropy(K.flatten(vae.inputs),
    #                                           K.flatten(vae.outputs))
    #
    # reconstruction_loss *= image_size * image_size
    # kl_loss = 1 + z_log_var - K.square(z_mean) - K.exp(z_log_var)
    # kl_loss = K.sum(kl_loss, axis=-1)
    # kl_loss *= -0.5
    # vae_loss = K.mean(reconstruction_loss + kl_loss)
    # vae.add_loss(vae_loss)
    # vae.compile(optimizer='rmsprop')
    # vae.summary()

    kl_loss = - 0.5 * K.sum(1 + z_log_sigma - K.square(z_mean) - K.exp(z_log_sigma), axis=-1)

    reconstruction_loss = binary_crossentropy(K.flatten(vae.inputs),
                                              K.flatten(vae.outputs))
    reconstruction_loss = mse(K.flatten(vae.inputs),
                              K.flatten(vae.outputs))
    reconstruction_loss *= image_size * image_size
    # kl_loss = 1 + z_log_sigma - K.square(z_mean) - K.exp(z_log_sigma)
    # kl_loss = K.sum(kl_loss, axis=-1)
    # kl_loss *= -0.5
    vae_loss = K.mean(reconstruction_loss + 0.1 * kl_loss)
    vae.add_loss(vae_loss)
    vae.compile(optimizer='rmsprop')
    encoder.add_loss(vae_loss)
    encoder.compile(optimizer='rmsprop')
    # decoder.compile(optimizer='rmsprop', loss=vae_loss)

    input_source = r'C:\Users\rossm\Documents\Data\Foraminifera\ForamA\project_2.xml'

    data_source = DataSource()
    data_source.use_mmap = False
    data_source.set_source(input_source, 40)
    data_source.load_dataset(img_size=(image_size, image_size),
                             prepro_type=None,
                             prepro_params=(255, 0, 1),
                             img_type='greyscale',
                             print_status=True,
                             dtype=np.float32)
    data_source.split(0.25, 0, 0)

    # if args.weights:
    #     vae.load_weights(args.weights)
    # else:
    #     # train the autoencoder
    vae.fit((data_source.train_images, data_source.train_images),
            epochs=200,
            batch_size=32,
            validation_data=(data_source.test_images, None))
        # vae.save_weights('vae_cnn_mnist.h5')

    # plot_results(models, data, batch_size=batch_size, model_name="vae_cnn")


















def base_cyclic_auto_encoder(input_shape,
                             filters=4,
                             blocks=4,
                             dropout=0.5,
                             dense=512,
                             conv_padding='same',
                             conv_activation='relu',
                             use_batch_norm=True,
                             global_pooling=None):

    inputs = Input(shape=input_shape)
    x = cyclic.CyclicSlice4()(inputs)
    for i in range(blocks):
        conv_filters = filters * 2 ** i
        # First layer
        x = Conv2D(conv_filters, (3, 3), padding=conv_padding, activation=None, kernel_initializer='he_normal')(x)
        x = Activation(conv_activation)(x)
        # Second layer
        x = Conv2D(conv_filters, (3, 3), padding=conv_padding, activation=None, kernel_initializer='he_normal')(x)
        x = Activation(conv_activation)(x)
        # Pool
        x = MaxPooling2D()(x)
        # Roll
        x = cyclic.CyclicRoll4()(x)
    # if global_pooling == 'avg':
    #     x = GlobalAveragePooling2D()(x)
    # elif global_pooling == 'max':
    #     x = GlobalMaxPooling2D()(x)
    # Dense layers
    encoded = x

    for i in range(blocks):
        conv_filters = filters * 2 ** i
        # First layer
        x = Conv2D(conv_filters, (3, 3), padding=conv_padding, activation=None, kernel_initializer='he_normal')(x)
        x = Activation(conv_activation)(x)
        # Second layer
        x = Conv2D(conv_filters, (3, 3), padding=conv_padding, activation=None, kernel_initializer='he_normal')(x)
        x = Activation(conv_activation)(x)
        # Pool
        x = MaxPooling2D()(x)
        # Roll
        x = cyclic.CyclicRoll4()(x)

    # x = cyclic.CyclicDensePool4(pool_op=tf.reduce_mean)(x)
    # x = Dropout(dropout)(x)
    # x = Dense(dense, activation='relu')(x)
    # x = cyclic.CyclicDensePool4(pool_op=tf.reduce_mean)(x)
    # x = Dense(nb_classes, activation='softmax')(x)

    model = Model(inputs, x, name='base_cyclic_auto_encoder')
    return model



import os
import collections
import tensorflow as tf
import tensorflow.keras as tfkeras
from miso.layers._common_blocks import ChannelSE
from miso.layers.cyclic import *
from tensorflow.keras.layers import GlobalMaxPooling2D, GlobalAveragePooling2D


# -------------------------------------------------------------------------
#   Helpers functions
# -------------------------------------------------------------------------

def handle_block_names(stage, block):
    name_base = 'stage{}_unit{}_'.format(stage + 1, block + 1)
    conv_name = name_base + 'conv'
    bn_name = name_base + 'bn'
    relu_name = name_base + 'relu'
    sc_name = name_base + 'sc'
    return conv_name, bn_name, relu_name, sc_name


def get_conv_params(**params):
    default_conv_params = {
        'kernel_initializer': 'he_uniform',
        'use_bias': False,
        'padding': 'valid',
    }
    default_conv_params.update(params)
    return default_conv_params


def get_bn_params(**params):
    axis = 3 if tfkeras.backend.image_data_format() == 'channels_last' else 1
    default_bn_params = {
        'axis': axis,
        'momentum': 0.99,
        'epsilon': 2e-5,
        'center': True,
        'scale': True,
    }
    default_bn_params.update(params)
    return default_bn_params


# -------------------------------------------------------------------------
#   Residual blocks
# -------------------------------------------------------------------------

def residual_conv_block(filters, stage, block, strides=(1, 1), attention=None, cut='pre', use_cyclic=False):
    """The identity block is the block that has no conv layer at shortcut.
    # Arguments
        input_tensor: input tensor
        kernel_size: default 3, the kernel size of
            middle conv layer at main path
        filters: list of integers, the filters of 3 conv layer at main path
        stage: integer, current stage label, used for generating layer names
        block: 'a','b'..., current block label, used for generating layer names
        cut: one of 'pre', 'post'. used to decide where skip connection is taken
    # Returns
        Output tensor for the block.
    """

    def layer(input_tensor):

        # get params and names of layers
        conv_params = get_conv_params()
        bn_params = get_bn_params()
        conv_name, bn_name, relu_name, sc_name = handle_block_names(stage, block)

        x = tfkeras.layers.BatchNormalization(name=bn_name + '1', **bn_params)(input_tensor)
        x = tfkeras.layers.Activation('relu', name=relu_name + '1')(x)

        # defining shortcut connection
        if cut == 'pre':
            shortcut = input_tensor
        elif cut == 'post':
            shortcut = tfkeras.layers.Conv2D(filters, (1, 1), name=sc_name, strides=strides, **conv_params)(x)
        else:
            raise ValueError('Cut type not in ["pre", "post"]')

        # continue with convolution layers
        x = tfkeras.layers.ZeroPadding2D(padding=(1, 1))(x)
        x = tfkeras.layers.Conv2D(filters, (3, 3), strides=strides, name=conv_name + '1', **conv_params)(x)
        # if use_cyclic:
        #     x = CyclicRoll4()(x)

        x = tfkeras.layers.BatchNormalization(name=bn_name + '2', **bn_params)(x)
        x = tfkeras.layers.Activation('relu', name=relu_name + '2')(x)
        x = tfkeras.layers.ZeroPadding2D(padding=(1, 1))(x)
        x = tfkeras.layers.Conv2D(filters, (3, 3), name=conv_name + '2', **conv_params)(x)
        if use_cyclic:
            x = CyclicRoll4()(x)

        # use attention block if defined
        if attention is not None:
            x = attention(x)

        # add residual connection
        x = tfkeras.layers.Add()([x, shortcut])
        return x

    return layer


def residual_bottleneck_block(filters, stage, block, strides=None, attention=None, cut='pre', use_cyclic=False):
    """The identity block is the block that has no conv layer at shortcut.
    # Arguments
        input_tensor: input tensor
        kernel_size: default 3, the kernel size of
            middle conv layer at main path
        filters: list of integers, the filters of 3 conv layer at main path
        stage: integer, current stage label, used for generating layer names
        block: 'a','b'..., current block label, used for generating layer names
        cut: one of 'pre', 'post'. used to decide where skip connection is taken
    # Returns
        Output tensor for the block.
    """

    def layer(input_tensor):

        # get params and names of layers
        conv_params = get_conv_params()
        bn_params = get_bn_params()
        conv_name, bn_name, relu_name, sc_name = handle_block_names(stage, block)

        x = tfkeras.layers.BatchNormalization(name=bn_name + '1', **bn_params)(input_tensor)
        x = tfkeras.layers.Activation('relu', name=relu_name + '1')(x)

        # defining shortcut connection
        if cut == 'pre':
            shortcut = input_tensor
        elif cut == 'post':
            shortcut = tfkeras.layers.Conv2D(filters * 4, (1, 1), name=sc_name, strides=strides, **conv_params)(x)
        else:
            raise ValueError('Cut type not in ["pre", "post"]')

        # continue with convolution layers
        x = tfkeras.layers.Conv2D(filters, (1, 1), name=conv_name + '1', **conv_params)(x)

        x = tfkeras.layers.BatchNormalization(name=bn_name + '2', **bn_params)(x)
        x = tfkeras.layers.Activation('relu', name=relu_name + '2')(x)
        x = tfkeras.layers.ZeroPadding2D(padding=(1, 1))(x)
        x = tfkeras.layers.Conv2D(filters, (3, 3), strides=strides, name=conv_name + '2', **conv_params)(x)

        x = tfkeras.layers.BatchNormalization(name=bn_name + '3', **bn_params)(x)
        x = tfkeras.layers.Activation('relu', name=relu_name + '3')(x)
        x = tfkeras.layers.Conv2D(filters * 4, (1, 1), name=conv_name + '3', **conv_params)(x)

        # use attention block if defined
        if attention is not None:
            x = attention(x)

        # add residual connection
        x = tfkeras.layers.Add()([x, shortcut])

        return x

    return layer


# -------------------------------------------------------------------------
#   Residual Model Builder
# -------------------------------------------------------------------------


def ResNetCyclic(model_params,
                 input_shape=None,
                 input_tensor=None,
                 include_top=True,
                 classes=1000,
                 weights='imagenet',
                 **kwargs):
    """Instantiates the ResNet, SEResNet architecture.
    Optionally loads weights pre-trained on ImageNet.
    Note that the data format convention used by the model is
    the one specified in your Keras config at `~/.keras/keras.json`.

    Args:
        include_top: whether to include the fully-connected
            layer at the top of the network.
        weights: one of `None` (random initialization),
              'imagenet' (pre-training on ImageNet),
              or the path to the weights file to be loaded.
        input_tensor: optional Keras tensor
            (i.e. output of `tfkeras.layers.Input()`)
            to use as image input for the model.
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False (otherwise the input shape
            has to be `(224, 224, 3)` (with `channels_last` data format)
            or `(3, 224, 224)` (with `channels_first` data format).
            It should have exactly 3 inputs channels.
        classes: optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.

    Returns:
        A Keras model instance.

    Raises:
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """

    if input_tensor is None:
        img_input = tfkeras.layers.Input(input_shape)
    else:
        if not tfkeras.backend.is_keras_tensor(input_tensor):
            img_input = tfkeras.layers.Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    # choose residual block type
    ResidualBlock = model_params.residual_block
    if model_params.attention:
        Attention = model_params.attention(**kwargs)
    else:
        Attention = None

    # get parameters for model layers
    no_scale_bn_params = get_bn_params(scale=False)
    bn_params = get_bn_params()
    conv_params = get_conv_params()
    init_filters = model_params.filters

    # resnet bottom
    if model_params.use_cyclic:
        x = CyclicSlice4()(img_input)
        # x = tfkeras.layers.BatchNormalization(name='bn_data', **no_scale_bn_params)(x)
    else:
        x = img_input
        # x = tfkeras.layers.BatchNormalization(name='bn_data', **no_scale_bn_params)(img_input)
    # x = tfkeras.layers.ZeroPadding2D(padding=(3, 3))(x)
    # x = tfkeras.layers.Conv2D(init_filters, (7, 7), strides=(2, 2), name='conv0', **conv_params)(x)

    # 2 layers of 3 x 3 to start!
    x = tfkeras.layers.ZeroPadding2D(padding=(1, 1))(x)
    x = tfkeras.layers.Conv2D(init_filters, (3, 3), strides=(1, 1), name='conv0a', **conv_params)(x)
    # if model_params.use_cyclic:
    #     x = CyclicRoll4()(x)
    x = tfkeras.layers.BatchNormalization(name='bn0', **bn_params)(x)
    x = tfkeras.layers.Activation('relu', name='relu0')(x)

    x = tfkeras.layers.ZeroPadding2D(padding=(1, 1))(x)
    x = tfkeras.layers.Conv2D(init_filters, (3, 3), strides=(1, 1), name='conv0b', **conv_params)(x)
    if model_params.use_cyclic:
        x = CyclicRoll4()(x)

    # x = tfkeras.layers.ZeroPadding2D(padding=(1, 1))(x)
    # x = tfkeras.layers.MaxPooling2D((3, 3), strides=(2, 2), padding='valid', name='pooling0')(x)

    # resnet body
    for stage, rep in enumerate(model_params.repetitions):
        for block in range(rep):

            filters = init_filters * (2 ** (stage + 1))

            # first block of first stage without strides because we have maxpooling before
            # if block == 0 and stage == 0:
            #     x = ResidualBlock(filters, stage, block, strides=(1, 1),
            #                       cut='post', attention=Attention)(x)

            if block == 0:
                x = ResidualBlock(filters, stage, block, strides=(2, 2),
                                  cut='post', attention=Attention)(x)

            else:
                x = ResidualBlock(filters, stage, block, strides=(1, 1),
                                  cut='pre', attention=Attention)(x)

    x = tfkeras.layers.BatchNormalization(name='bn1', **bn_params)(x)
    x = tfkeras.layers.Activation('relu', name='relu1')(x)

    if model_params.global_pooling == 'avg':
        x = GlobalAveragePooling2D()(x)
    elif model_params.global_pooling == 'max':
        x = GlobalMaxPooling2D()(x)

    x = tfkeras.layers.Flatten()(x)
    x = CyclicDensePool4(pool_op=tf.reduce_mean)(x)
    x = tfkeras.layers.Dropout(0.5)(x)
    x = tfkeras.layers.Dense(512, activation='relu')(x)
    x = tfkeras.layers.Dense(classes, activation='softmax')(x)

    # # resnet top
    # if include_top:
    #     x = tfkeras.layers.GlobalAveragePooling2D(name='pool1')(x)
    #     x = tfkeras.layers.Dense(classes, name='fc1')(x)
    #     x = tfkeras.layers.Activation('softmax', name='softmax')(x)
    # Ensure that the model takes into account any potential predecessors of `input_tensor`.
    if input_tensor is not None:
        inputs = tfkeras.keras_utils.get_source_inputs(input_tensor)
    else:
        inputs = img_input

    # Create model.
    model = tfkeras.models.Model(inputs, x)
    return model


ResnetModelParameters = collections.namedtuple(
    'ModelParams',
    ['model_name', 'filters', 'repetitions', 'residual_block', 'attention', 'use_cyclic', 'global_pooling']
)

MODELS_PARAMS = {
    'resnet18': ResnetModelParameters('resnet18', 64, (2, 2, 2, 2), residual_conv_block, None, use_cyclic=True, global_pooling=None),
    'resnet34': ResnetModelParameters('resnet34', 64, (3, 4, 6, 3), residual_conv_block, None, use_cyclic=True, global_pooling=None),
    'resnet50': ResnetModelParameters('resnet50', 64, (3, 4, 6, 3), residual_bottleneck_block, None, use_cyclic=True, global_pooling=None),
    'resnet101': ResnetModelParameters('resnet101', 64, (3, 4, 23, 3), residual_bottleneck_block, None,
                                       use_cyclic=True, global_pooling=None),
    'resnet152': ResnetModelParameters('resnet152', 64, (3, 8, 36, 3), residual_bottleneck_block, None,
                                       use_cyclic=True, global_pooling=None),
    'seresnet18': ResnetModelParameters('seresnet18', 64, (2, 2, 2, 2), residual_conv_block, ChannelSE,
                                        use_cyclic=True, global_pooling=None),
    'seresnet34': ResnetModelParameters('seresnet34', 64, (3, 4, 6, 3), residual_conv_block, ChannelSE,
                                        use_cyclic=True, global_pooling=None),
}



import glob
import os
from collections import OrderedDict
import pandas as pd
import platform


def list_directory(source_dir):
    return sorted(glob.glob(os.path.join(source_dir, "*")))


def parse_directory(source_dir, verbose=True, skip='~', has_classes=True, sub_directories=[]):
    """
    Parses a directory consisting of subdirectories named by class and stores the image filenames in a dictionary
    :param source_dir: The directory to parse
    :param verbose: Print out the progress
    :param skip: If a subdirectory starts with this character it will be skipped ('_' and '.' are skipped always, unless skip is None)
    :param has_classes: The files are arranged in subsdirectories by class name
    :return: Dictionary with class names for the keys and lists of filenames for the values
    """
    if has_classes:
        sub_dirs = sorted(glob.glob(os.path.join(source_dir, "*")))
    else:
        sub_dirs = [source_dir]
    filenames = OrderedDict()
    if verbose:
        print("Parsing directory {}".format(source_dir))
    for sub_dir in sub_dirs:
        if os.path.isdir(sub_dir) is False:
            continue
        # Get the directory name
        sub_name = os.path.basename(sub_dir)
        # Skip directories starting with ~
        if sub_name.startswith(skip) or \
            sub_name.startswith('_') or \
            sub_name.startswith('.'):
            continue
        # Get the files
        if verbose:
            print("- {}".format(sub_name), end='')
        files = []
        for ext in ["*.jpg", "*.jpeg", "*.png", "*.tif", "*.tiff", "*.bmp"]:
            sub_files1 = sorted(glob.glob(os.path.join(sub_dir, *sub_directories, ext)))
            files.extend(sub_files1)
            if platform.system() == 'Linux':
                sub_files2 = sorted(glob.glob(os.path.join(sub_dir, *sub_directories, ext.upper())))
                files.extend(sub_files2)
        # Add to dictionary
        if verbose:
            print(" ({} files)".format(len(files)))
        if has_classes:
            filenames[sub_name] = files
        else:
            filenames['null'] = files
    return filenames


def flatten_to_list(filenames_dict: dict):
    filenames = []
    for value in filenames_dict.values():
        filenames.extend(value)
    return filenames


def parse_csv(csv_file, source_dir, file_idx=0, cls_idx=1, cls_label_idx=2, verbose=True):
    if verbose:
        print("Parsing csv file {} for directory {}".format(csv_file, source_dir))
    df = pd.read_csv(csv_file)
    num_classes = np.max(df.iloc[:, cls_idx]) + 1
    cls_labels = [df.loc[df.iloc[:, cls_idx] == i].iloc[0, cls_label_idx] for i in range(num_classes)]
    filenames = OrderedDict()
    for i in range(num_classes):
        if verbose:
            print("- {}".format(cls_labels[i]), end='')
        names = df.loc[df.iloc[:, cls_idx] == i].iloc[:, file_idx]
        paths = [os.path.join(source_dir, n) for n in names]
        filenames[cls_labels[i]] = paths
        if verbose:
            print(" ({} files)".format(len(paths)))
    return filenames


if __name__ == "__main__":
    from mml.data.filenames_dataset import FilenamesDataset
    import numpy as np

    result = parse_csv(r"D:\Datasets\Weeds\labels.csv",
                       r"D:\Datasets\Weeds\DeepWeeds")

    # fs = FilenamesDataset(r"C:\data\SeagrassFrames")
    # fs.split(0.2)
    # arr = np.zeros(len(fs.train_filenames))
    # load_images(fs.train_filenames, arr, None)


"""
Creates and trains a generic network
"""
import time
import datetime
from collections import OrderedDict

from miso.stats.mislabelling import find_and_save_mislabelled
from miso.archive.datasource import DataSource
from miso.archive.generators import *
from miso.utils.wave import *
from miso.training.adaptive_learning_rate import AdaptiveLearningRateScheduler
from miso.training.training_result import TrainingResult
from miso.stats.confusion_matrix import *
from miso.stats.training import *
from archive.augmentation import *
from miso.deploy.saving import freeze, convert_to_inference_mode
from miso.deploy.model_info import ModelInfo
from miso.models.model_factory import *


def train_image_classification_model(params: dict, data_source: DataSource = None):
    # Make both backends use the same session
    K.clear_session()

    intro()

    # Params -----------------------------------------------------------------------------------------------------------
    name = params.get('name')
    description = params.get('description')

    # Network
    cnn_type = params.get('type')

    # Input
    img_height = params.get('img_height')
    img_width = params.get('img_width')
    img_channels = params.get('img_channels')

    # Training
    batch_size = params.get('batch_size')
    max_epochs = params.get('max_epochs')
    alr_epochs = params.get('alr_epochs')
    alr_drops = params.get('alr_drops')

    # Input data
    input_dir = params.get('input_source')
    data_min_count = params.get('data_min_count')
    data_split = params.get('data_split')
    data_split_offset = params.get('data_split_offset')
    seed = params.get('seed')

    # Output
    output_dir = params.get('save_dir')

    # Data -------------------------------------------------------------------------------------------------------------
    # print("@Loading images...")
    if img_channels == 3:
        color_mode = 'rgb'
    else:
        if cnn_type.endswith('tl'):
            color_mode = 'greyscale3'
            params['img_channels'] = 3
        else:
            color_mode = 'greyscale'
    print('Color mode: {}'.format(color_mode))

    if data_source is None:
        data_source = DataSource()
        data_source.use_mmap = params['use_mmap']
        data_source.set_source(input_dir,
                               data_min_count,
                               mapping=params['class_mapping'],
                               min_count_to_others=params['data_map_others'],
                               mmap_directory=params['mmap_directory'])
        data_source.load_dataset(img_size=(img_height, img_width),
                                 prepro_type=None,
                                 prepro_params=(255, 0, 1),
                                 img_type=color_mode,
                                 print_status=True)
    data_source.split(data_split, data_split_offset, seed)

    if params['use_class_weights'] is True:
        params['class_weights'] = data_source.get_class_weights()
        print("@Class weights are {}".format(params['class_weights']))
    else:
        params['class_weights'] = None
    params['num_classes'] = data_source.num_classes

    if cnn_type.endswith('tl'):

        # mnist = tf.keras.datasets.mnist
        # (x_train, y_train), (x_test, y_test) = mnist.load_data()
        # x_train, x_test = x_train / 255.0, x_test / 255.0
        # model = tf.keras.models.Sequential([
        #     tf.keras.layers.Flatten(input_shape=(28, 28)),
        #     tf.keras.layers.Dense(128, activation='relu'),
        #     tf.keras.layers.Dropout(0.2),
        #     tf.keras.layers.Dense(10)
        # ])
        # loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
        # model.compile(optimizer='adam',
        #               loss=loss_fn,
        #               metrics=['accuracy'])
        # model.fit(x_train, y_train, epochs=5)
        # model.evaluate(x_test, y_test, verbose=2)
        # probability_model = tf.keras.Sequential([
        #     model,
        #     tf.keras.layers.Softmax()
        # ])

        start = time.time()
        # Create Vectors -----------------------------------------------------------------------------------------------
        # Note that the images are scaled internally in the network to match the expected preprocessing
        # print(time.time())
        # model_head.predict(data_source.train_images[0:1024])
        # print(time.time())
        # model_head.predict(data_source.train_images[0:1024])
        # print(time.time())
        # test = data_source.train_images[0:1024].copy()
        # model_head.predict(test)
        # print(time.time())
        # model_head.predict(data_source.train_images[0:1024])
        # print(time.time())
        # test = data_source.train_images[0:1024].copy()
        # model_head.predict(test)
        # print(time.time())

        # Generate vectors
        model_head = generate_tl_head(params)
        print("@Calculating train vectors")
        t = time.time()
        train_vector = model_head.predict(data_source.train_images)
        print("!{}s elapsed".format(time.time() - t))
        print("@Calculating test vectors")
        t = time.time()
        test_vector = model_head.predict(data_source.test_images)
        print("!{}s elapsed".format(time.time() - t))
        # Clear
        K.clear_session()

        # print(train_vector.dtype)
        # print(test_vector.dtype)

        # Generate vectors (random!)
        # train_vector = np.random.random(size=[data_source.train_images.shape[0], 2048])
        # test_vector = np.random.random(size=[data_source.test_images.shape[0], 2048])

        # train_vector = []
        # test_vector = []
        # step = 64
        #
        # for i in range(0, len(data_source.train_images), step):
        #     train_vector.append(model_head.predict(data_source.train_images[i:i+step]))
        #     print("@Calculating train vectors - {} of {}".format(i, len(data_source.train_images)))
        # train_vector = np.concatenate(train_vector, axis=0)
        #
        # for i in range(0, len(data_source.test_images), step):
        #     test_vector.append(model_head.predict(data_source.test_images[i:i + step]))
        #     print("@Calculating test vectors - {} of {}".format(i, len(data_source.test_images)))
        # test_vector = np.concatenate(test_vector, axis=0)

        data_source.train_vectors = train_vector
        data_source.test_vectors = test_vector

        # Augmentation -------------------------------------------------------------------------------------------------
        # No augmentation - there is a bug in the batch normalisation layer for tensorflow v1.xx where the mean and variance
        # are still calculated even when the layer is set to not trainable. This means the vectors produced are not the
        # vary according to the batch. For augmentation we need to include the ResNet network (with its batch normalisation
        # layers) in the graph, and because of this bug, the training performance is poor.

        # Generator ----------------------------------------------------------------------------------------------------
        # No generator needed

        # Model --------------------------------------------------------------------------------------------------------
        print("@Generating tail")
        # Get  tail
        model_tail = generate_tl_tail(params, [train_vector.shape[1], ])
        model_tail.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])

        # Training -----------------------------------------------------------------------------------------------------
        # alr_cb = AdaptiveLearningRateScheduler(nb_epochs=alr_epochs,
        #                                        nb_drops=alr_drops)
        # print("@Training")
        # if data_split > 0:
        #     validation_data = (test_vector, data_source.test_onehots)
        # else:
        #     validation_data = None
        # history = model_tail.fit(train_vector,
        #                          data_source.train_onehots,
        #                          validation_data=validation_data,
        #                          epochs=max_epochs,
        #                          batch_size=batch_size,
        #                          shuffle=True,
        #                          verbose=0,
        #                          class_weight=params['class_weights'],
        #                          callbacks=[alr_cb])
        # end = time.time()
        # training_time = end - start
        # print("@Training time: {}s".format(training_time))
        # time.sleep(3)

        # Generator ----------------------------------------------------------------------------------------------------
        train_gen = tf_vector_generator(train_vector,
                                        data_source.train_onehots,
                                        batch_size)
        test_gen = tf_vector_generator(test_vector,
                                       data_source.test_onehots,
                                       batch_size)

        # Training -----------------------------------------------------------------------------------------------------
        alr_cb = AdaptiveLearningRateScheduler(nb_epochs=alr_epochs,
                                               nb_drops=alr_drops)
        print("@Training")
        if data_split > 0:
            validation_data = test_gen
        else:
            validation_data = None
        # log_dir = "C:\\logs\\profile\\" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        # tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1, profile_batch=3)
        history = model_tail.fit_generator(
            train_gen,
            steps_per_epoch=math.ceil(len(train_vector) // batch_size),
            validation_data=validation_data,
            validation_steps=math.ceil(len(test_vector) // batch_size),
            epochs=max_epochs,
            verbose=0,
            shuffle=False,
            max_queue_size=1,
            class_weight=params['class_weights'],
            callbacks=[alr_cb])
        end = time.time()
        training_time = end - start
        print("@Training time: {}s".format(training_time))
        time.sleep(3)

        # Generator ----------------------------------------------------------------------------------------------------
        # Now we be tricky and join the trained dense layers to the resnet model to create a model that accepts images
        # as input
        model_head = generate_tl_head(params)
        outputs = model_tail(model_head.output)
        model = Model(model_head.input, outputs)
        model.summary()

        # Vector -------------------------------------------------------------------------------------------------------
        vector_model = generate_vector(model, params)

    else:
        # Model --------------------------------------------------------------------------------------------------------
        print("@Generating model")
        start = time.time()
        model = generate(params)

        # Augmentation -------------------------------------------------------------------------------------------------
        if params['aug_rotation'] is True:
            rotation_range = [0, 360]
        else:
            rotation_range = None

        def augment(x):
            return augmentation_complete(x,
                                         rotation=rotation_range,
                                         gain=params['aug_gain'],
                                         gamma=params['aug_gamma'],
                                         zoom=params['aug_zoom'],
                                         gaussian_noise=params['aug_gaussian_noise'],
                                         bias=params['aug_bias'])

        if params['use_augmentation'] is True:
            augment_fn = augment
        else:
            augment_fn = None

        # Generator ----------------------------------------------------------------------------------------------------
        train_gen = tf_augmented_image_generator(data_source.train_images,
                                                 data_source.train_onehots,
                                                 batch_size,
                                                 augment_fn)
        test_gen = image_generator(data_source.test_images,
                                   data_source.test_onehots,
                                   batch_size)

        # Training -----------------------------------------------------------------------------------------------------

        alr_cb = AdaptiveLearningRateScheduler(nb_epochs=alr_epochs,
                                               nb_drops=alr_drops)
        print("@Training")
        if data_split > 0:
            validation_data = test_gen
        else:
            validation_data = None
        history = model.fit_generator(
            train_gen,
            steps_per_epoch=math.ceil(len(data_source.train_images) // batch_size),
            validation_data=validation_data,
            validation_steps=math.ceil(len(data_source.test_images) // batch_size),
            epochs=max_epochs,
            verbose=0,
            shuffle=False,
            max_queue_size=1,
            class_weight=params['class_weights'],
            callbacks=[alr_cb])
        end = time.time()
        training_time = end - start
        print("@Training time: {}s".format(training_time))
        time.sleep(3)

        # Vector -------------------------------------------------------------------------------------------------------
        vector_model = generate_vector(model, params)

    # Graphs -----------------------------------------------------------------------------------------------------------
    print("@Generating results")
    if data_split > 0:
        # Calculate test set scores
        y_true = data_source.test_cls
        y_prob = model.predict(data_source.test_images)
        y_pred = y_prob.argmax(axis=1)
    else:
        y_true = np.asarray([])
        y_prob = np.asarray([])
        y_pred = np.asarray([])

    # Inference time
    max_count = np.min([1000, len(data_source.images)])
    to_predict = np.copy(data_source.images[0:max_count])

    inf_times = []
    for i in range(3):
        start = time.time()
        model.predict(to_predict)
        end = time.time()
        diff = (end - start) / max_count * 1000
        inf_times.append(diff)
        print("@Calculating inference time {}/10: {:.3f}ms".format(i + 1, diff))
    inference_time = np.median(inf_times)

    # Store results
    result = TrainingResult(params,
                            history,
                            y_true,
                            y_pred,
                            y_prob,
                            data_source.cls_labels,
                            training_time,
                            inference_time)

    # Save the results
    now = datetime.datetime.now()
    save_dir = os.path.join(output_dir, "{0}_{1:%Y%m%d-%H%M%S}".format(name, now))
    os.makedirs(save_dir, exist_ok=True)

    # Plot the graphs
    # plot_model(model, to_file=os.path.join(save_dir, "model_plot.pdf"), show_shapes=True)
    if data_split > 0:
        plot_loss_vs_epochs(history)
        plt.savefig(os.path.join(save_dir, "loss_vs_epoch.pdf"))
        plot_accuracy_vs_epochs(history)
        plt.savefig(os.path.join(save_dir, "accuracy_vs_epoch.pdf"))
        plot_confusion_accuracy_matrix(y_true, y_pred, data_source.cls_labels)
        plt.savefig(os.path.join(save_dir, "confusion_matrix.pdf"))
        plt.close('all')

    if params['save_mislabeled'] is True:
        print("@Estimating mislabeled")
        vectors = vector_model.predict(data_source.images)
        find_and_save_mislabelled(data_source.images,
                                  vectors,
                                  data_source.cls,
                                  data_source.cls_labels,
                                  data_source.get_short_filenames(),
                                  save_dir,
                                  11)

    # Save model -------------------------------------------------------------------------------------------------------
    print("@Saving model")
    # Convert if necessary to fix TF batch normalisation issues
    model = convert_to_inference_mode(model, lambda: generate(params))
    vector_model = generate_vector(model, params)

    # Generate description
    if description is None:
        description = "{}: {} model trained on data from {} ({} images in {} classes).\n" \
                      "Accuracy: {:.1f} (P: {:.1f}, R: {:.1f}, F1 {:.1f})".format(
            name,
            cnn_type,
            input_dir,
            len(data_source.data_df),
            len(data_source.cls_labels),
            result.accuracy * 100,
            result.mean_precision() * 100,
            result.mean_recall() * 100,
            result.mean_f1_score() * 100)

    # Create model info with all the parameters
    inputs = OrderedDict()
    inputs["image"] = model.inputs[0]
    outputs = OrderedDict()
    outputs["pred"] = model.outputs[0]
    outputs["vector"] = vector_model.outputs[0]
    info = ModelInfo(name,
                     description,
                     cnn_type,
                     now,
                     "frozen_model.pb",
                     params,
                     inputs,
                     outputs,
                     input_dir,
                     data_source.cls_labels,
                     data_source.cls_counts,
                     "rescale",
                     [255, 0, 1],
                     result.accuracy,
                     result.precision,
                     result.recall,
                     result.f1_score,
                     result.support,
                     result.epochs[-1],
                     training_time,
                     params['data_split'],
                     inference_time)

    # Freeze and save graph
    if params['save_model'] is not None:
        freeze(model,
               os.path.join(save_dir, "model"),
               info)

    # Save info
    info.save(os.path.join(save_dir, "model", "network_info.xml"))

    print("@Deleting temporary files")
    data_source.delete_memmap_files(del_split=True, del_source=params['delete_mmap_files'])

    wave()

    print("@Complete")
    return model, vector_model, data_source, result

# tensorboard_cb = TensorBoard(log_dir='./tensorboard',
#                              histogram_freq=0,
#                              batch_size=32,
#                              write_graph=True,
#                              write_grads=False,
#                              write_images=False,
#                              embeddings_freq=0,
#                              embeddings_layer_names=None,
#                              embeddings_metadata=None,
#                              embeddings_data=None,
#                              update_freq='epoch')



from tensorflow.keras.layers import Layer, InputSpec
from tensorflow.keras import initializers
from tensorflow.keras import regularizers
from tensorflow.keras import constraints
from tensorflow.keras import backend as K

# from tensorflow.keras.utils.generic_utils import get_custom_objects


class GroupNormalization(Layer):
    """Group normalization layer

    Group Normalization divides the channels into groups and computes within each group
    the mean and variance for normalization. GN's computation is independent of batch sizes,
    and its accuracy is stable in a wide range of batch sizes

    # Arguments
        groups: Integer, the number of groups for Group Normalization.
        axis: Integer, the axis that should be normalized
            (typically the features axis).
            For instance, after a `Conv2D` layer with
            `data_format="channels_first"`,
            set `axis=1` in `BatchNormalization`.
        epsilon: Small float added to variance to avoid dividing by zero.
        center: If True, add offset of `beta` to normalized tensor.
            If False, `beta` is ignored.
        scale: If True, multiply by `gamma`.
            If False, `gamma` is not used.
            When the next layer is linear (also e.g. `nn.relu`),
            this can be disabled since the scaling
            will be done by the next layer.
        beta_initializer: Initializer for the beta weight.
        gamma_initializer: Initializer for the gamma weight.
        beta_regularizer: Optional regularizer for the beta weight.
        gamma_regularizer: Optional regularizer for the gamma weight.
        beta_constraint: Optional constraint for the beta weight.
        gamma_constraint: Optional constraint for the gamma weight.

    # Input shape
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    # Output shape
        Same shape as input.

    # References
        - [Group Normalization](https://arxiv.org/abs/1803.08494)
    """

    def __init__(self,
                 groups=32,
                 axis=-1,
                 epsilon=1e-5,
                 center=True,
                 scale=True,
                 beta_initializer='zeros',
                 gamma_initializer='ones',
                 beta_regularizer=None,
                 gamma_regularizer=None,
                 beta_constraint=None,
                 gamma_constraint=None,
                 **kwargs):
        super(GroupNormalization, self).__init__(**kwargs)
        self.supports_masking = True
        self.groups = groups
        self.axis = axis
        self.epsilon = epsilon
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)
        self.beta_regularizer = regularizers.get(beta_regularizer)
        self.gamma_regularizer = regularizers.get(gamma_regularizer)
        self.beta_constraint = constraints.get(beta_constraint)
        self.gamma_constraint = constraints.get(gamma_constraint)

    def build(self, input_shape):
        dim = input_shape[self.axis]

        if dim is None:
            raise ValueError('Axis ' + str(self.axis) + ' of '
                             'input tensor should have a defined dimension '
                             'but the layer received an input with shape ' +
                             str(input_shape) + '.')

        if dim < self.groups:
            raise ValueError('Number of groups (' + str(self.groups) + ') cannot be '
                             'more than the number of channels (' +
                             str(dim) + ').')

        if dim % self.groups != 0:
            raise ValueError('Number of groups (' + str(self.groups) + ') must be a '
                             'multiple of the number of channels (' +
                             str(dim) + ').')

        self.input_spec = InputSpec(ndim=len(input_shape),
                                    axes={self.axis: dim})
        shape = (dim,)

        if self.scale:
            self.gamma = self.add_weight(shape=shape,
                                         name='gamma',
                                         initializer=self.gamma_initializer,
                                         regularizer=self.gamma_regularizer,
                                         constraint=self.gamma_constraint)
        else:
            self.gamma = None
        if self.center:
            self.beta = self.add_weight(shape=shape,
                                        name='beta',
                                        initializer=self.beta_initializer,
                                        regularizer=self.beta_regularizer,
                                        constraint=self.beta_constraint)
        else:
            self.beta = None
        self.built = True

    def call(self, inputs, **kwargs):
        input_shape = K.int_shape(inputs)
        tensor_input_shape = K.shape(inputs)

        # Prepare broadcasting shape.
        reduction_axes = list(range(len(input_shape)))
        del reduction_axes[self.axis]
        broadcast_shape = [1] * len(input_shape)
        broadcast_shape[self.axis] = input_shape[self.axis] // self.groups
        broadcast_shape.insert(1, self.groups)

        reshape_group_shape = K.shape(inputs)
        group_axes = [reshape_group_shape[i] for i in range(len(input_shape))]
        group_axes[self.axis] = input_shape[self.axis] // self.groups
        group_axes.insert(1, self.groups)

        # reshape inputs to new group shape
        group_shape = [group_axes[0], self.groups] + group_axes[2:]
        group_shape = K.stack(group_shape)
        inputs = K.reshape(inputs, group_shape)

        group_reduction_axes = list(range(len(group_axes)))
        group_reduction_axes = group_reduction_axes[2:]

        mean = K.mean(inputs, axis=group_reduction_axes, keepdims=True)
        variance = K.var(inputs, axis=group_reduction_axes, keepdims=True)

        inputs = (inputs - mean) / (K.sqrt(variance + self.epsilon))

        # prepare broadcast shape
        inputs = K.reshape(inputs, group_shape)
        outputs = inputs

        # In this case we must explicitly broadcast all parameters.
        if self.scale:
            broadcast_gamma = K.reshape(self.gamma, broadcast_shape)
            outputs = outputs * broadcast_gamma

        if self.center:
            broadcast_beta = K.reshape(self.beta, broadcast_shape)
            outputs = outputs + broadcast_beta

        outputs = K.reshape(outputs, tensor_input_shape)

        return outputs

    def get_config(self):
        config = {
            'groups': self.groups,
            'axis': self.axis,
            'epsilon': self.epsilon,
            'center': self.center,
            'scale': self.scale,
            'beta_initializer': initializers.serialize(self.beta_initializer),
            'gamma_initializer': initializers.serialize(self.gamma_initializer),
            'beta_regularizer': regularizers.serialize(self.beta_regularizer),
            'gamma_regularizer': regularizers.serialize(self.gamma_regularizer),
            'beta_constraint': constraints.serialize(self.beta_constraint),
            'gamma_constraint': constraints.serialize(self.gamma_constraint)
        }
        base_config = super(GroupNormalization, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape


# get_custom_objects().update({'GroupNormalization': GroupNormalization})


if __name__ == '__main__':
    from tensorflow.keras.layers import Input
    from tensorflow.keras.models import Model
    ip = Input(shape=(None, None, 4))
    x = GroupNormalization(groups=2, axis=-1, epsilon=0.1)(ip)
    model = Model(ip, x)
    model.summary()



import tensorflow as tf
import os
from zipfile import ZipFile
import hashlib


def download_images(origin, directory):
    # Fix URLs
    # - OneDrive
    if origin.startswith("https://1drv.ms/"):
        origin = origin.replace("https://1drv.ms/", "https://1drv.ws/")

    hash = hashlib.md5(origin.encode()).hexdigest()[:10]
    directory = os.path.join(directory, hash)
    zip_path = os.path.join(directory, "download.zip")
    if os.path.exists(directory) is False or os.path.exists(zip_path) is True:
        os.makedirs(directory, exist_ok=True)
        outp = tf.keras.utils.get_file(
            "download.zip",
            origin,
            untar=False,
            md5_hash=None,
            file_hash=None,
            cache_subdir=directory,
            hash_algorithm='auto',
            extract=True,
            archive_format='auto',
            cache_dir=None)
        print(outp)
        # Create a ZipFile Object and load sample.zip in it
        with ZipFile(zip_path, 'r') as zip_obj:
            # Get list of files names in zip
            paths = [path for path in zip_obj.namelist() if not path.startswith("_")]
            dir_name = paths[0].replace(',', '/').replace(',', '\\')
            folder_path = os.path.join(directory, dir_name)
        os.remove(os.path.join(directory, "download.zip"))
    else:
        paths = [path for path in os.listdir(directory) if not path.startswith("_")]
        folder_path = os.path.join(directory, paths[0])
        print("Already downloaded at: {}".format(folder_path))
    return folder_path





import math
from collections import namedtuple

from miso.models.transfer_learning import *
from miso.models.base_cyclic import *
from miso.models.resnet_cyclic import *
from classification_models.tfkeras import Classifiers, ModelsFactory
from miso.training.parameters import MisoParameters

try:
    from tensorflow.keras.applications.efficientnet import *
except ImportError:
    pass


def generate(tp: MisoParameters):
    #
    # Transfer learning
    #
    # Uses any of the available models in keras applications
    if tp.cnn.id.endswith('tl'):
        parts = tp.cnn.id.split("_")
        if parts[0] in TRANSFER_LEARNING_PARAMS.keys():
            model_head, model_tail = generate_tl(tp.cnn.id, tp.dataset.num_classes, tp.cnn.img_shape)
            return combine_tl(model_head, model_tail)
        else:
            raise ValueError("The CNN type {} is not supported, valid CNNs are {}".format(parts[0], TRANSFER_LEARNING_PARAMS.keys()))

    #
    # Full network training
    #
    # Base Cyclic - custom network created at CEREGE specifically for foraminifera by adding cyclic layers
    if tp.cnn.id.startswith("base_cyclic"):
        model = base_cyclic(input_shape=tp.cnn.img_shape,
                            nb_classes=tp.dataset.num_classes,
                            filters=tp.cnn.filters,
                            blocks=tp.cnn.blocks,
                            dropout=0.5,
                            dense=512,
                            conv_activation=tp.cnn.activation,
                            use_batch_norm=tp.cnn.use_batch_norm,
                            global_pooling=tp.cnn.global_pooling)
    # ResNet Cyclic - custom network created at CEREGE specifically for foraminifera by adding cyclic layers
    elif tp.cnn.id.startswith("resnet_cyclic"):
        blocks = int(math.log2(tp.cnn.img_shape[0]) - 2)
        blocks -= 1  # Resnet has one block to start with already
        resnet_params = ResnetModelParameters('resnet_cyclic',
                                    tp.cnn.filters,
                                    [1 for i in range(blocks)],
                                    residual_conv_block,
                                    None,
                                    use_cyclic=True,
                                    global_pooling=tp.cnn.global_pooling)
        model = ResNetCyclic(resnet_params, tp.cnn.img_shape, None, True, tp.dataset.num_classes)
    # EfficientNet
    elif tp.cnn.id.lower().startswith(("efficientnet")):
        if tp.cnn.id.lower() == "efficientnetb0":
            model_fn = EfficientNetB0
        elif tp.cnn.id.lower() == "efficientnetb1":
            model_fn = EfficientNetB1
        elif tp.cnn.id.lower() == "efficientnetb2":
            model_fn = EfficientNetB2
        elif tp.cnn.id.lower() == "efficientnetb3":
            model_fn = EfficientNetB3
        elif tp.cnn.id.lower() == "efficientnetb4":
            model_fn = EfficientNetB4
        elif tp.cnn.id.lower() == "efficientnetb5":
            model_fn = EfficientNetB5
        elif tp.cnn.id.lower() == "efficientnetb6":
            model_fn = EfficientNetB6
        elif tp.cnn.id.lower() == "efficientnetb7":
            model_fn = EfficientNetB7
        model = model_fn(weights=None,
                         input_shape=tp.cnn.img_shape,
                         classes=tp.dataset.num_classes)
    # ResNet, SEResNet, DenseNet and others from qubvel's image-classifiers python package
    elif tp.cnn.id in ModelsFactory().models.keys():
        classifier, preprocess_input = Classifiers.get(tp.cnn.id)
        model = classifier(input_shape=tp.cnn.img_shape,
                           weights=None,
                           classes=tp.dataset.num_classes)
    else:
        raise ValueError(
            "The CNN type {} is not supported, valid CNNs are base_cyclic, resnet_cyclic, efficientnetb[0-7] and {}".format(tp.cnn.id, ModelsFactory().models.keys()))
    model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])
    return model


def generate_tl_head(cnn_type, img_shape):
    model_head = head(cnn_type, input_shape=img_shape)
    return model_head


def generate_tl_tail(num_classes, input_shape):
    model_tail = tail(num_classes, input_shape)
    return model_tail


def generate_tl(cnn_type, num_classes, img_shape):
    model_head = generate_tl_head(cnn_type, img_shape)
    model_tail = tail(num_classes, [model_head.layers[-1].output.shape[-1], ])
    return model_head, model_tail

def combine_tl(model_head, model_tail):
    return Model(inputs=model_head.input, outputs=model_tail.call(model_head.output))


def generate_vector(model, cnn_type):
    if cnn_type.endswith("tl"):
        vector_tensor = model.get_layer(index=-2).get_output_at(1)
        vector_model = Model(model.inputs, vector_tensor)
    elif cnn_type.startswith("base_cyclic"):
        vector_layer = model.get_layer(index=-2)
        vector_model = Model(model.inputs, vector_layer.output)
    elif cnn_type.startswith("resnet_cyclic"):
        vector_layer = model.get_layer(index=-2)
        vector_model = Model(model.inputs, vector_layer.output)
    elif cnn_type.startswith("resnet") or cnn_type.startswith("seresnet"):
        vector_layer = model.get_layer(index=-3)
        vector_model = Model(model.inputs, vector_layer.output)
    elif cnn_type.startswith("vgg") or cnn_type.startswith("densenet"):
        vector_layer = model.get_layer(index=-2)
        vector_model = Model(model.inputs, vector_layer.output)
    else:
        vector_layer = model.get_layer(index=-2)
        vector_model = Model(model.inputs, vector_layer.output)
    return vector_model



"""
Creates and trains a generic network
"""
import warnings

from miso.training.helpers import predict_in_batches
from miso.training.train_full_network import train_full_network
from miso.training.train_transfer_learning import train_transfer_learning

warnings.simplefilter(action="ignore", category=FutureWarning)

import time
import datetime
from collections import OrderedDict
import keras.backend as K
from sklearn.manifold import TSNE

from miso.data.training_dataset import TrainingDataset
from miso.stats.embedding import plot_embedding
from miso.stats.mislabelling import find_and_save_mislabelled
from miso.training.training_result import TrainingResult
from miso.stats.confusion_matrix import *
from miso.stats.training import *
from miso.deploy.saving import (
    save_frozen_model_tf2,
    convert_to_inference_mode_tf2,
    load_from_xml,
    save_model_as_onnx,
)
from miso.deploy.model_info import ModelInfo
from miso.models.factory import *

import matplotlib.pyplot as plt
import pandas as pd


def train_image_classification_model(tp: MisoParameters):
    now = datetime.datetime.now()

    # Hack to make RTX cards work
    physical_devices = tf.config.list_physical_devices("GPU")
    for device in physical_devices:
        tf.config.experimental.set_memory_growth(device, True)

    K.clear_session()

    # Clean the training parameters
    tp.sanitise()

    print("+---------------------------------------------------------------------------+")
    print("| MISO Particle Classification Library                                      |")
    print("+---------------------------------------------------------------------------+")
    print("| Stable version:                                                           |")
    print("| pip install miso                                                          |")
    print("| Development version:                                                      |")
    print("| pip install git+http://www.github.com/microfossil/particle-classification |")
    print("+---------------------------------------------------------------------------+")
    print("Tensorflow version: {}".format(tf.__version__))
    print("-" * 80)
    print("Train information:")
    print("- name: {}".format(tp.name))
    print("- description: {}".format(tp.description))
    print("- CNN type: {}".format(tp.cnn.id))
    print("- image type: {}".format(tp.cnn.img_type))
    print("- image shape: {}".format(tp.cnn.img_shape))
    print()

    # Load data
    ds = TrainingDataset(
        tp.dataset.source,
        tp.cnn.img_shape,
        tp.cnn.img_type,
        tp.dataset.min_count,
        tp.dataset.map_others,
        tp.dataset.train_split,
        tp.dataset.val_split,
        tp.dataset.random_seed,
        tp.dataset.memmap_directory,
    )
    ds.load()
    tp.dataset.num_classes = ds.num_classes

    print("Split:")
    print(f"- train: {ds.train_split}, {len(ds.train_idx)} images")
    print(f"- val: {ds.test_split}, {len(ds.test_idx)} images")

    # Create save locations
    save_dir = os.path.join(
        tp.output.save_dir, "{0}_{1:%Y%m%d-%H%M%S}".format(tp.name, now)
    )
    os.makedirs(save_dir, exist_ok=True)

    # ------------------------------------------------------------------------------
    # Training
    # ------------------------------------------------------------------------------
    # Transfer learning
    if tp.training.use_transfer_learning or tp.cnn.id.endswith("_tl"):
        model, history = train_transfer_learning(tp, ds)
    # Full network
    else:
        model, history = train_full_network(tp, ds, save_dir)

    # ------------------------------------------------------------------------------
    # Evaluation
    # ------------------------------------------------------------------------------
    print("-" * 80)
    print("Evaluating model")

    # Accuracy
    if tp.dataset.val_split > 0:
        y_true = ds.cls[ds.test_idx]
        gen = ds.test_generator(tp.training.batch_size, shuffle=False, one_shot=True)
        y_prob = model.predict(gen.create())
        y_pred = y_prob.argmax(axis=1)
    else:
        y_true = np.asarray([])
        y_prob = np.asarray([])
        y_pred = np.asarray([])

    # Inference time
    print("- calculating inference time:", end="")
    max_count = np.min([128, len(ds.images.data)])
    inf_times = []
    for i in range(3):
        gen = ds.images.create_generator(
            tp.training.batch_size,
            idxs=np.arange(max_count),
            shuffle=False,
            one_shot=True,
        )
        start = time.time()
        model.predict(gen.create())
        end = time.time()
        diff = (end - start) / max_count * 1000
        inf_times.append(diff)
        print(" {:.3f}ms".format(diff), end="")
    inference_time = np.median(inf_times)
    print(", median: {}".format(inference_time))

    # Store results
    # - fix to make key same for tensorflow 1 and 2
    if "accuracy" in history.history:
        history.history["acc"] = history.history.pop("accuracy")
    if "val_accuracy" in history.history:
        history.history["val_acc"] = history.history.pop("val_accuracy")
    training_time = history.history["training_time"]
    result = TrainingResult(
        tp,
        history,
        y_true,
        y_pred,
        y_prob,
        ds.cls_labels,
        training_time,
        inference_time,
    )
    print("- accuracy {:.2f}".format(result.accuracy * 100))
    print("- mean precision {:.2f}".format(result.mean_precision * 100))
    print("- mean recall {:.2f}".format(result.mean_recall * 100))

    # ------------------------------------------------------------------------------
    # Save results
    # ------------------------------------------------------------------------------
    if tp.description is None:
        tp.description = (
            "{}: {} model trained on data from {} ({} images in {} classes).\n"
            "Accuracy: {:.1f} (P: {:.1f}, R: {:.1f}, F1 {:.1f})".format(
                tp.name,
                tp.cnn.id,
                tp.dataset.source,
                len(ds.filenames.filenames),
                len(ds.cls_labels),
                result.accuracy * 100,
                result.mean_precision * 100,
                result.mean_recall * 100,
                result.mean_f1_score * 100,
            )
        )

    # Create model info with all the parameters
    inputs = OrderedDict()
    inputs["image"] = model.inputs[0]
    outputs = OrderedDict()
    outputs["pred"] = model.outputs[0]
    # outputs["vector"] = vector_model.outputs[0]
    info = ModelInfo(
        tp.name,
        tp.description,
        tp.cnn.id,
        now,
        "frozen_model.pb",
        tp,
        inputs,
        outputs,
        tp.dataset.source,
        ds.cls_labels,
        ds.filenames.cls_counts,
        "rescale",
        [255, 0, 1],
        result.accuracy,
        result.precision,
        result.recall,
        result.f1_score,
        result.support,
        result.epochs[-1],
        training_time,
        tp.dataset.val_split,
        inference_time,
    )

    # ------------------------------------------------------------------------------
    # Cleanlab health report
    # ------------------------------------------------------------------------------
    print("-" * 80)
    print("Health report")
    import cleanlab

    report = cleanlab.dataset.health_summary(y_true, y_prob, class_names=ds.cls_labels)
    with open(os.path.join(save_dir, "health_summary.txt"), "w") as fp:
        fp.write("Dataset health report\n")
        fp.write("\n")
        fp.write("This is calculated using the test data\n")
        fp.write("\n" + "-" * 80 + "\n")
        fp.write(f"Overall label heath: {report['overall_label_health_score']}")
        # fp.write("-" * 80 + "\n")
        # fp.write("Joint probabilities\n")
        # fp.write(report["joint"])
        fp.write("\n" + "-" * 80 + "\n")
        fp.write("Classes by label quality\n")
        fp.write(report["classes_by_label_quality"].to_string())
        fp.write("\n" + "-" * 80 + "\n")
        fp.write("Overlapping classes\n")
        fp.write(report["overlapping_classes"].to_string())

    # ------------------------------------------------------------------------------
    # Plots
    # ------------------------------------------------------------------------------
    # Plot the graphs
    # plot_model(model, to_file=os.path.join(save_dir, "model_plot.pdf"), show_shapes=True)
    print("-" * 80)
    print("Plotting")
    if tp.dataset.val_split > 0:
        print("- loss")
        plot_loss_vs_epochs(history)
        plt.savefig(os.path.join(save_dir, "loss_vs_epoch.pdf"))
        print("- accuracy")
        plot_accuracy_vs_epochs(history)
        plt.savefig(os.path.join(save_dir, "accuracy_vs_epoch.pdf"))
        print("- confusion matrix")
        plot_confusion_accuracy_matrix(y_true, y_pred, ds.cls_labels)
        plt.savefig(os.path.join(save_dir, "confusion_matrix.pdf"))
        plt.close("all")

    # Mislabelled
    vector_model = generate_vector_from_model(model, tp)
    if tp.output.save_mislabeled is True:
        print("- mislabeled")
        print("- calculating vectors... ", end="")
        gen = ds.images.create_generator(
            tp.training.batch_size, shuffle=False, one_shot=True
        )
        vectors = vector_model.predict(gen.create())
        print("{} total".format(len(vectors)))
        find_and_save_mislabelled(
            ds.images.data,
            vectors,
            ds.cls,
            ds.cls_labels,
            [os.path.basename(f) for f in ds.filenames.filenames],
            save_dir,
            11,
        )

    # t-SNE
    print("- t-SNE (1024 vectors max)")
    print("- calculating vectors... ", end="")
    idxs = np.random.choice(
        np.arange(len(ds.images.data)),
        np.min((1024, len(ds.images.data))),
        replace=False,
    )
    gen = ds.images.create_generator(
        tp.training.batch_size, idxs=idxs, shuffle=False, one_shot=True
    )
    vec_subset = vector_model.predict(gen.create())
    X = TSNE(n_components=2).fit_transform(vec_subset)
    plot_embedding(X, ds.cls[idxs], ds.num_classes)
    plt.savefig(os.path.join(save_dir, "tsne.pdf"))
    cls_info = pd.DataFrame({"index": range(ds.num_classes), "label": ds.cls_labels})
    cls_info.to_csv(os.path.join(save_dir, "legend.csv"), sep=";")

    # ------------------------------------------------------------------------------
    # Save model (has to be last thing it seems)
    # ------------------------------------------------------------------------------
    print("-" * 80)
    print("Saving model")
    # Convert if necessary to fix TF batch normalisation issues

    # Freeze and save graph
    if tp.output.save_model is not None:
        inference_model = convert_to_inference_mode_tf2(model, lambda: generate(tp))
        # tf.saved_model.save(inference_model, os.path.join(os.path.join(save_dir, "model_keras")))
        frozen_func = save_frozen_model_tf2(
            inference_model, os.path.join(save_dir, "model_tf2"), "frozen_model.pb"
        )
        info.protobuf = "frozen_model.pb"
        info.inputs["image"] = frozen_func.inputs[0]
        info.outputs["pred"] = frozen_func.outputs[0]
        info.save(os.path.join(save_dir, "model_tf2", "network_info.xml"))

        save_model_as_onnx(
            inference_model,
            inference_model.inputs[0].name,
            [
                None,
            ]
            + tp.cnn.img_shape,
            os.path.join(os.path.join(save_dir, "model_onnx")),
        )
        info.protobuf = "model.onnx"
        info.inputs["image"] = inference_model.inputs[0]
        info.outputs["pred"] = inference_model.outputs[0]
        info.save(os.path.join(save_dir, "model_onnx", "network_info.xml"))

    # Save parameters
    tp.save(os.path.join(save_dir, "training_parameters.json"))

    # ------------------------------------------------------------------------------
    # Confirm model save
    # ------------------------------------------------------------------------------
    if tp.output.save_model is not None and tp.dataset.val_split > 0:
        print("-" * 80)
        print("Validate saved model")
        y_pred_old = y_pred
        y_true = ds.cls[ds.test_idx]
        gen = ds.test_generator(32, shuffle=False, one_shot=True)
        y_prob = []
        model, img_size, cls_labels = load_from_xml(
            os.path.join(save_dir, "model_tf2", "network_info.xml")
        )
        for b in iter(gen.to_tfdataset()):
            y_prob.append(model(b[0]).numpy())
        y_prob = np.concatenate(y_prob, axis=0)
        y_pred = y_prob.argmax(axis=1)
        acc = accuracy_score(y_true, y_pred)
        p, r, f1, _ = precision_recall_fscore_support(y_true, y_pred)
        print(
            "Saved model on test set: acc {:.2f}, prec {:.2f}, rec {:.2f}, f1 {:.2f}".format(
                acc, np.mean(p), np.mean(r), np.mean(f1)
            )
        )
        acc = accuracy_score(y_pred_old, y_pred)
        if acc == 1.0:
            print("Overlap: {:.2f}% - PASSED".format(acc * 100))
        else:
            print("Overlap: {:.2f}% - FAILED".format(acc * 100))

    # ------------------------------------------------------------------------------
    # Clean up
    # ------------------------------------------------------------------------------
    print("- cleaning up")
    ds.release()
    print("- complete")
    print("-" * 80)
    print()
    return model, vector_model, ds, result



import tensorflow as tf
from tensorflow.keras.layers import Dense, Dropout, Flatten, Conv2D, MaxPooling2D, Input, Activation, \
                                    GlobalMaxPooling2D, GlobalAveragePooling2D, Lambda, DepthwiseConv2D
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.models import Model

import numpy as np

from miso.layers import cyclic

def base_cyclic(input_shape,
                nb_classes,
                filters=4,
                blocks=None,
                dropout=0.5,
                dense=512,
                conv_padding='same',
                conv_activation='relu',
                use_batch_norm=True,
                global_pooling=None,
                use_depthwise_conv=True):
    # Number of blocks
    if blocks is None:
        blocks = int(np.log2(input_shape[0]) - 2)
    inputs = Input(shape=input_shape)
    x = cyclic.CyclicSlice4()(inputs)
    # Convolution blocks
    for i in range(blocks):
        conv_filters = filters * 2 ** i
        for j in range(2):
            x = Conv2D(conv_filters, (3, 3), padding=conv_padding, activation=None, kernel_initializer='he_normal')(x)
            if use_batch_norm is True:
                x = BatchNormalization()(x)
            x = Activation(conv_activation)(x)
        x = MaxPooling2D()(x)
        x = cyclic.CyclicRoll4()(x)
    if global_pooling == 'avg':
        x = GlobalAveragePooling2D()(x)
    elif global_pooling == 'max':
        x = GlobalMaxPooling2D()(x)
    # Dense layers
    x = Flatten()(x)
    x = cyclic.CyclicDensePool4(pool_op=tf.reduce_mean)(x)
    x = Dropout(dropout)(x)
    x = Dense(dense, activation='relu')(x)
    x = Dense(nb_classes, activation='softmax')(x)
    # Return model
    model = Model(inputs, x, name='base_cyclic')
    return model


def mirror_cyclic(input_shape,
                nb_classes,
                filters=4,
                blocks=4,
                dropout=0.5,
                dense=512,
                conv_padding='same',
                conv_activation='relu',
                use_batch_norm=True,
                global_pooling=None):

    default_bn_params = {
        'axis': 3,
        'momentum': 0.99,
        'epsilon': 2e-5,
        'center': True,
        'scale': True,
    }

    inputs = Input(shape=input_shape)
    x = cyclic.CyclicSlice4()(inputs)
    for i in range(blocks):
        conv_filters = filters * 2 ** i
        # First layer
        x = Conv2D(conv_filters, (3, 3), padding=conv_padding, activation=None, kernel_initializer='he_normal')(x)
        if use_batch_norm is True:
            # x = GroupNormalization(conv_filters)(x)
            x = BatchNormalization()(x)
            # x = LayerNormalization()(x)
            # x = BatchInstanceNormalisation()(x)
            # x = Lambda(lambda x: batch_instance_norm(x, scope="bin_{}_0".format(i)))(x)
            # x = Lambda((lambda x: tf.layers.batch_normalization(x, training=K.learning_phase())))(x)
        xa = Activation(conv_activation)(x)
        xb = Activation(conv_activation)(-x)
        x = tf.stack([xa, xb], 4)
        x = tf.reduce_max(x, axis=-1)
        # Second layer
        x = Conv2D(conv_filters, (3, 3), padding=conv_padding, activation=None, kernel_initializer='he_normal')(x)
        if use_batch_norm is True:
            # x = GroupNormalization(conv_filters)(x)
            x = BatchNormalization()(x)
            # x = LayerNormalization()(x)
            # x = BatchInstanceNormalisation()(x)
            # x = Lambda(lambda x: batch_instance_norm(x, scope="bin_{}_1".format(i)))(x)
            # x = Lambda((lambda x: tf.layers.batch_normalization(x, training=K.learning_phase())))(x)
        # x = Activation(conv_activation)(x)
        xa = Activation(conv_activation)(x)
        xb = Activation(conv_activation)(-x)
        x = tf.stack([xa, xb], 4)
        x = tf.reduce_max(x, axis=-1)
        # Pool
        x = MaxPooling2D()(x)
        # Roll
        x = cyclic.CyclicRoll4()(x)
    if global_pooling == 'avg':
        x = GlobalAveragePooling2D()(x)
    elif global_pooling == 'max':
        x = GlobalMaxPooling2D()(x)

    # Dense layers
    x = Flatten()(x)
    x = cyclic.CyclicDensePool4(pool_op=tf.reduce_mean)(x)
    x = Dropout(dropout)(x)
    x = Dense(dense, activation='relu')(x)
    # x = cyclic.CyclicDensePool4(pool_op=tf.reduce_mean)(x)
    x = Dense(nb_classes, activation='softmax')(x)

    model = Model(inputs, x, name='base_cyclic')
    return model




import os
import random
from pathlib import Path

import cv2
import matplotlib.pyplot as plt
import numpy as np
import onnx
import pandas as pd
import skimage.io
from PIL import Image
from tqdm import tqdm

from miso.data.image_utils import load_image
from miso.deploy.model_info import ModelInfo
from miso.deploy.saving import load_frozen_model_tf2, load_from_xml, load_onnx_from_xml
import tensorflow as tf

from miso.inference.morphology import MorphologyProcessor


def get_image_paths_and_samples(base_path, sample_name="unknown"):
    base_path = Path(base_path)
    image_paths = []
    samples = []
    for subdir in base_path.iterdir():
        if subdir.is_dir():
            for file in subdir.rglob('*'):
                if file.suffix.lower() in ('.png', '.jpg', '.jpeg', '.bmp', '.gif', '.tiff', '.tif'):
                    image_paths.append(str(file))
                    samples.append(subdir.name)
        else:
            file = subdir
            if file.suffix.lower() in ('.png', '.jpg', '.jpeg', '.bmp', '.gif', '.tiff', '.tif'):
                image_paths.append(str(file))
                samples.append(sample_name)
    return image_paths, samples


def classify_folder(model_info_path,
                    images_path,
                    output_path,
                    batch_size,
                    sample_name="unknown",
                    unsure_threshold=0.0):
    model, img_size, img_type, labels = load_from_xml(model_info_path)

    # Create a dataset of image paths
    image_paths, sample_names = get_image_paths_and_samples(images_path, sample_name)
    image_dataset = tf.data.Dataset.from_tensor_slices(image_paths)

    # Image loading functions
    def load_and_preprocess_image(image_path):
        def _load_image(image_path):
            image_path = image_path.numpy().decode('utf-8')
            image = load_image(image_path, img_size, img_type)
            image = tf.convert_to_tensor(image, dtype=tf.float32)
            image = image / 255.0
            return image

    # Map using the image dataset
    image_dataset = image_dataset.map(load_and_preprocess_image, num_parallel_calls=tf.data.AUTOTUNE)

    # Batch the dataset
    image_dataset = image_dataset.batch(batch_size)

    # Run predictions
    predictions = []
    idxs = []
    cls = []
    scores = []
    for batch in tqdm(image_dataset):
        preds = model(batch).numpy()
        batch_idxs = np.argmax(preds, axis=1)
        batch_labels = [labels[idx] for idx in batch_idxs]
        batch_scores = np.max(preds, axis=1)
        predictions.extend(preds)
        idxs.extend(batch_idxs)
        cls.extend(batch_labels)
        scores.extend(batch_scores)

    idxs = [idx if score > unsure_threshold else -1 for idx, score in zip(idxs, scores)]
    cls = [cls if score > unsure_threshold else "unsure" for cls, score in zip(cls, scores)]

    df = pd.DataFrame({
        "filename": image_paths,
        "short_filename": [Path(f).relative_to(images_path) for f in image_paths],
        "sample": sample_names,
        "class_index": idxs,
        "class": cls,
        "score": scores
    })
    df.to_csv(output_path, index=False)


import onnxruntime as rt


def segment_folder(model_info_path,
                   images_path,
                   output_path,
                   batch_size,
                   sample_name="unknown",
                   threshold=0.5,
                   save_contours=False):

    sess, img_size, img_type, _ = load_onnx_from_xml(model_info_path)

    image_paths, sample_names = get_image_paths_and_samples(images_path, sample_name)
    image_dataset = tf.data.Dataset.from_tensor_slices(image_paths)

    def load_and_preprocess_image(image_path):
        def _load_image(image_path):
            image_path = image_path.numpy().decode('utf-8')
            image = load_image(image_path, img_size, img_type)
            image = tf.convert_to_tensor(image, dtype=tf.float32)
            image = image / 255.0
            return image
        image = tf.py_function(_load_image, [image_path], tf.float32)
        image.set_shape([img_size[0], img_size[1], 1])
        return image

    image_dataset = image_dataset.map(load_and_preprocess_image, num_parallel_calls=tf.data.AUTOTUNE)
    image_dataset = image_dataset.batch(batch_size)

    contours = []
    for batch in tqdm(image_dataset):
        batch = batch.numpy()
        masks = sess.run(None, {sess.get_inputs()[0].name: batch})[0]
        # masks = model(batch).numpy()
        for mask in masks:
            mask = (mask > threshold).astype(np.uint8)  # Threshold the mask
            cnts, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            if cnts:
                # Find the largest contour based on area
                areas = [cv2.contourArea(cnt) for cnt in cnts]
                largest_contour = cnts[np.argmax(areas)]
                contours.append(largest_contour)
            else:
                contours.append(None)

    # Loop through the images again, loading greyscale and then calculating the morphology
    datas = []
    idx = 0
    for image_path, sample_name, contour in tqdm(zip(image_paths, sample_names, contours)):
        idx += 1
        image = Image.open(image_path)
        greyscale = np.array(image.convert('L'), dtype=np.float32)

        # Make square
        border_pixels = np.concatenate(
            [greyscale[0, :],
             greyscale[-1, :],
             greyscale[:, 0],
             greyscale[:, -1]]
        )
        border_mean = np.median(border_pixels)
        orig_size = greyscale.shape[:2]
        diff = abs(orig_size[0] - orig_size[1])
        d1 = diff // 2
        d2 = diff - d1
        padding = ((d1, d2), (0, 0)) if orig_size[0] < orig_size[1] else ((0, 0), (d1, d2))
        greyscale = np.pad(greyscale, padding, mode='constant', constant_values=border_mean)

        # Get max size
        max_size = max(greyscale.shape)

        # Get scale factor
        scale_factor = max_size / img_size[0]

        # Scale contour
        if contour is not None:
            contour = contour * scale_factor

        m = MorphologyProcessor.calculate_morphology(None, greyscale, contour)

        data = {}

        data["filename"] = image_path
        data["sample"] = sample_name
        data.update(m.__dict__)

        datas.append(data)

        # Save the contour
        if contour is not None and save_contours:
            # Save path
            output_path = Path(output_path)
            contour_image_path = output_path.parent / "contours" / sample_name
            contour_image_path.mkdir(parents=True, exist_ok=True)
            rgb = cv2.cvtColor(greyscale, cv2.COLOR_GRAY2RGB).astype(np.uint8)
            cv2.drawContours(rgb, [contour.astype(int)], -1, (255, 0, 0), 1)
            skimage.io.imsave(contour_image_path / f"{Path(image_path).stem}_contour.png", rgb, check_contrast=False)

        if idx % 1000 == 0:
            df = pd.DataFrame(datas)
            df.to_csv(output_path, index=False)


if __name__ == "__main__":
    pass
    # def _load_image(image_path):
    #     image_path = image_path
    #     image = Image.open(image_path)
    #     greyscale = np.array(image.convert('L'), dtype=np.float32)[..., np.newaxis]
    #     greyscale = greyscale[:,50:,:]
    #     border_pixels = np.concatenate(
    #         [greyscale[0, :, 0],
    #          greyscale[-1, :, 0],
    #          greyscale[:, 0, 0],
    #          greyscale[:, -1, 0]]
    #     )
    #     border_mean = np.median(border_pixels)
    #     orig_size = greyscale.shape[:2]
    #     diff = abs(orig_size[0] - orig_size[1])
    #     d1 = diff // 2
    #     d2 = diff - d1
    #     padding = ((d1, d2), (0, 0), (0, 0)) if orig_size[0] < orig_size[1] else ((0, 0), (d1, d2), (0, 0))
    #     image = np.pad(greyscale, padding, mode='constant', constant_values=border_mean)
    #     image = tf.image.resize(image, [224, 224])
    #     image = image / 255.0
    #     return image
    #
    # # Test load image
    # image_path = r"C:\Users\ross.marchant\data\_Zooscan_centering_NOprocess\_inputs\_images\train_processed\crust_amphipoda\013-crus_amphipoda.png"
    # image = _load_image(image_path)
    #
    # plt.imshow(image[..., 0], cmap='gray')
    # plt.show()

    # os.environ['TF_CUDNN_DETERMINISTIC'] = '1'
    #
    # segment_folder(
    #     r"C:\Users\ross.marchant\code\Microfossil\particle-trieur\target\classes\trained_networks\plankton_segmenter\model_info.xml",
    #     r"C:\Users\ross.marchant\data\_Zooscan_centering_NOprocess\_inputs\_images\train_processed",
    #     r"F:\morphology.csv",
    #     64,
    #     sample_name="unknown",
    #     threshold=0.8,
    #     save_contours=True)




import argparse
from flask import Flask, request, flash
import numpy as np
from miso.data.image_utils import load_image
from miso.deploy.saving import load_from_xml
import sys
import pandas as pd
import tensorflow as tf

app = Flask(__name__)


@app.route('/')
def index():
    return '''
    <!doctype html>
    <title>MISO Classification Server</title>
    <h1>MISO Classification Server</h1>
    <h2>Classification</h2>
    <p>Use the <a href="/file">file</a> end point to classify an image</p>
    <p>Use the <a href="/count">count</a> end point to get the class counts</p>
    '''


@app.route('/count', methods=['GET'])
def counts():
    if request.method == 'GET':
        if 'format' in request.args and request.args['format'] == 'csv':
            return app_df.to_csv()
        else:
            return app_df.to_json()


@app.route('/file', methods=['GET'])
def classify_file():
    global app_df
    if request.method == 'GET':
        # check if the post request has the file part
        print("Classification request:")
        if 'filename' not in request.args:
            print("- no filename, serving webpage instead...")
            return '''
            <!doctype html>
            <title>Import file</title>
            <h1>Import and classify a particle image</h1>
            <p>Enter the full file path to import the image and classify it</p>
            <form method="GET">
                    File path:</br>
            <input type=text name=filename>
            </br>
                    Sample:</br>
            <input type=text name=sample>
            </br>
                    Index 1:</br>
            <input type=text name=index1>
            </br>
                    Index 2:</br>
            <input type=text name=index2>
            </br>
                    Resolution (pixels/mm):</br>
            <input type=text name=resolution>
                    </br>
            <input type=submit value=Upload>
            </form>'''
        else:
            filename = request.args['filename']
            sample = request.args['sample']
            index1 = request.args['index1']
            index2 = request.args['index2']
            resolution = request.args['resolution']
            print(f" - filename:   {filename}")
            print(f" - sample:     {sample}")
            print(f" - index1:     {index1}")
            print(f" - index2:     {index2}")
            print(f" - resolution: {resolution}")

            if filename == "":
                flash('{"error":"No filename was entered"}')
                return
            try:
                # Load image
                if app_img_size[2] == 1:
                    img_type = 'rgm'
                else:
                    img_type = 'greyscale'
                im = load_image(filename, app_img_size, img_type)

                # Classify
                result = app_model(tf.convert_to_tensor(im[np.newaxis, :, :, :], dtype=tf.float32)).numpy()[0]

                # Predictions
                idx = np.argmax(result)
                code = app_cls_labels[idx]
                score = np.max(result)

                # Format response
                result_str = "{{\"code\":\"{}\", \"score\":{}}}".format(code, score)
                if len(app_df) == 0:
                    app_df = app_df.append(pd.Series([sample] + [0] * len(app_cls_labels), index=app_df.columns), ignore_index=True)
                    app_df.set_index('sample', inplace=True)
                elif app_df.index.str.contains(sample) is False:
                    app_df = app_df.append(pd.Series([0] * len(app_cls_labels), name=sample, index=app_df.columns), ignore_index=False)
                app_df.loc[sample][idx] = app_df.loc[sample][idx] + 1
                print("Results\n - code: {}\n - score: {}".format(filename, code, score))
                return result_str
            except Exception as e:
                # Erros
                exc_type, exc_obj, exc_tb = sys.exc_info()
                print("Exception on line: {}".format(exc_tb.tb_lineno))
                print(e)
                result_str = "{{\"error\":\"{}\"}}".format(e)
                return result_str


if __name__ == '__main__':
    parser = argparse.ArgumentParser("MISO Classification Server")
    parser.add_argument("-i", "--info", required=True, help="CNN network information file")
    parser.add_argument("-p", "--port", required=True, help="Server port")
    args = parser.parse_args()

    app_model, app_img_size, app_cls_labels = load_from_xml(args.info)
    app_df = pd.DataFrame(columns=['sample'] + app_cls_labels)
    print("MISO Classification Server - port {}".format(args.port))
    print("--------------------------")
    print("Labels:")
    print(app_cls_labels)
    app.run(debug=False, port=args.port)





"""
https://towardsdatascience.com/enhancing-the-power-of-softmax-for-image-classification-4f8f85141739

https://github.com/christk1/MSH_tensorflow_keras/blob/master/mnist.py

x = tf.keras.layers.Flatten()(x)
x = tf.keras.layers.Dense(params.embedding_size, activation='relu', name='feats0')(x)
x = tf.keras.layers.Dense(params.embedding_size, name='features')(x)
aux_input = tf.keras.Input(shape=(params.num_classes,))
predictions = MSoftMaxLayer(n_classes=params.num_classes, m=params.m, s=params.s, name='MSoftMaxLayer')(
    [x, aux_input])
model = tf.keras.models.Model(inputs=[base_model.input, aux_input], outputs=predictions)


REMOVE RELU FROM FINAL LAYER!
"""
import tensorflow as tf


class MSoftMaxLayer(tf.keras.layers.Layer):

    def __init__(self, n_classes, m=0.5, s=64., **kwargs):
        self.num_classes = n_classes
        self.m = m
        self.s = s
        super(MSoftMaxLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        # Create a weight variable for this layer.
        self.kernel = self.add_weight(name='kernel',
                                      shape=(input_shape[0][-1], self.num_classes),
                                      initializer=tf.random_normal_initializer(stddev=0.01),
                                      trainable=True)
        # input_shape[0] contains the batch
        super(MSoftMaxLayer, self).build(input_shape)  # Be sure to call this at the end

    def call(self, inps):
        # emb (N x embs), labels (N, 10) = labels one hot
        emb, labels = inps

        # normalize feature
        emb = tf.nn.l2_normalize(emb, axis=1) * self.s  # (n, 512)
        # normalize weights
        W = tf.nn.l2_normalize(self.kernel, axis=0)  # (512, 10)
        fc7 = tf.matmul(emb, W)  # n x 10

        # pick elements along axis 1
        zy = tf.reduce_max(input_tensor=tf.multiply(fc7, labels), axis=1)  # (n, 1)

        cos_t = zy / self.s
        t = tf.acos(cos_t)
        body = tf.cos(t + self.m)
        new_zy = body * self.s
        diff = new_zy - zy
        diff = tf.expand_dims(diff, 1)
        body = tf.multiply(labels, diff)
        fc7 = fc7 + body
        fc7 = tf.nn.softmax(fc7)

        return fc7

    def get_config(self):
        config = super(MSoftMaxLayer, self).get_config()
        config.update({'num_classes': self.num_classes, 'm': self.m, 's': self.s})
        return config

    def compute_output_shape(self, input_shape):
        return None, self.num_classes


"""
Training parameters
"""
import copy

from pathlib import Path
from typing import Union, List, Optional
from dataclasses import dataclass, field
import re
from marshmallow_dataclass import class_schema

from miso.models.keras_models import KERAS_MODEL_PARAMETERS
from miso.utils.compact_json import CompactJSONEncoder


def default_field(obj):
    return field(default_factory=lambda: copy.copy(obj))

@dataclass
class BaseParameters(object):
    class Meta:
        ordered = True

    def dumps(self):
        return class_schema(self.__class__)().dumps(self, indent=4, cls=CompactJSONEncoder)

    def save(self, path: str):
        with open(path, "w") as fp:
            metadata = class_schema(self.__class__)().dumps(self, indent=4)
            fp.write(metadata)

    @classmethod
    def loads(cls, json):
        instance = class_schema(cls)().loads(json)
        return instance
    @classmethod
    def load(cls, path: Union[Path, str]):
        if isinstance(path, str):
            path = Path(path)
        instance = class_schema(cls)().loads(path.read_text())
        return instance


@dataclass
class ModelParameters(BaseParameters):
    # Common values
    id: str = "base_cyclic"
    img_shape: List[int] = field(default=(128, 128, 1))
    img_type: str = "greyscale"

    # Transfer learning cyclic layers
    use_cyclic: bool = False
    use_cyclic_gain: bool = False

    # Parameters for custom networks
    filters: Optional[int] = 4
    blocks: Optional[int] = None
    dense: Optional[int] = None
    use_batch_norm: bool = True
    global_pooling: bool = None
    activation: str = "relu"
    use_asoftmax: bool = False


@dataclass
class TrainingParameters(BaseParameters):
    batch_size: int = 64
    max_epochs: int = 10000
    alr_epochs: int = 10
    alr_drops: int = 4
    monitor_val_loss: bool = False
    use_class_weights: bool = True
    use_class_undersampling: bool = False
    use_augmentation: bool = True
    use_transfer_learning: bool = False
    transfer_learning_augmentation_factor: int = 0


@dataclass
class OptimizerParameters(BaseParameters):
    name: str = "adam"
    learning_rate: float = 0.001
    momentum: float = 0.9
    decay: float = 0.0
    nesterov: bool = False


@dataclass
class DatasetParameters(BaseParameters):
    num_classes: Optional[int] = None
    source: Optional[str] = None
    min_count: Optional[int] = 10
    train_split: Optional[float] = None
    val_split: Optional[float] = 0.2
    map_others: bool = False
    random_seed: int = 0
    memmap_directory: str = None


@dataclass
class AugmentationParameters(BaseParameters):
    rotation: List[float] = field(default=(0, 360))
    gain: Optional[List[float]] = field(default=(0.8, 1, 1.2))
    gamma: Optional[List[float]] = field(default=(0.5, 1, 2))
    bias: Optional[List[float]] = None
    zoom: Optional[List[float]] = field(default=(0.9, 1, 1.1))
    gaussian_noise: Optional[List[float]] = field(default=(0.01, 0.1))
    random_crop: Optional[List[int]] = None
    orig_img_shape: Optional[List[int]] = field(default=(256, 256, 3))


@dataclass
class OutputParameters(BaseParameters):
    save_dir: str = None
    save_model: bool = True
    save_mislabeled: bool = False


@dataclass
class MisoParameters(BaseParameters):
    name: str = ""
    description: str = ""
    cnn: ModelParameters = default_field(ModelParameters())
    dataset: DatasetParameters = default_field(DatasetParameters())
    training: TrainingParameters = default_field(TrainingParameters())
    augmentation: AugmentationParameters = default_field(AugmentationParameters())
    output: OutputParameters = default_field(OutputParameters())
    optimizer: OptimizerParameters = default_field(OptimizerParameters())

    def sanitise(self):
        if self.name == "":
            self.name = self.dataset.source[:64] + "_" + self.cnn.id
            self.name = re.sub('[^A-Za-z0-9]+', '-', self.name)
        if self.cnn.img_shape is None:
            if self.cnn.id.endswith("_tl"):
                shape = KERAS_MODEL_PARAMETERS[self.cnn.id.split('_')[0]].default_input_shape
            else:
                if self.cnn.id.startswith("base_cyclic") or self.cnn.id.startswith("resnet_cyclic"):
                    shape = [128, 128, 3]
                else:
                    shape = [224, 224, 3]
                if self.cnn.img_type == 'rgb':
                    shape[2] = 3
                    self.augmentation.orig_img_shape[2] = 3
                else:
                    shape[2] = 1
                    self.augmentation.orig_img_shape[2] = 3
            self.cnn.img_shape = shape
        elif self.cnn.id.startswith("base_cyclic") or self.cnn.id.startswith("resnet_cyclic"):
            pass
        else:
            self.cnn.img_shape[2] = 3


def get_default_shape(cnn_type):
    if cnn_type.endswith("_tl"):
        return KERAS_MODEL_PARAMETERS[cnn_type.split('_')[0]].default_input_shape
    else:
        return [224, 224, None]


if __name__ == "__main__":
    m = MisoParameters()
    print(m.dumps())










import numpy as np
import skimage.io as skio

from miso.data.dataset import DatasetBase
from miso.data.image_loader import ParallelImageLoader
from miso.data.image_utils import resize_transform, resize_with_pad_transform, null_transform
from miso.data.tf_generator import TFGenerator


class ImageDataset(DatasetBase):
    def __init__(self,
                 filenames,
                 cls=None,
                 transform_fn=None,
                 transform_args=None,
                 img_size=None,
                 memmap_directory=None,
                 overwrite_memmap=False,
                 unique_id=None,
                 dtype=np.uint8):
        self.filenames = filenames
        self.cls = cls
        self.transform_fn = transform_fn
        self.transform_args = transform_args
        self.unique_id = unique_id

        # Pre-made transforms
        if self.transform_fn == 'resize':
            self.transform_fn = resize_transform
        elif self.transform_fn == 'resize_with_pad':
            self.transform_fn = resize_with_pad_transform

        if self.transform_fn is None:
            self.transform_fn = null_transform
            self.transform_args = [0]

        # Get dataset unique identification hash
        super().__init__(memmap_directory=memmap_directory, overwrite_memmap=overwrite_memmap, dtype=dtype)
        self.hash_data = ["ImageDataset", self.filenames, str(self.dtype), self.unique_id]

        print('-' * 80)
        print("Loading images")
        print("- id: {}".format(self.get_hash_id()))
        if memmap_directory is None:
            print("- stored in RAM")
        else:
            print("- stored on disk at {}".format(self.memmap_file))

        if img_size is None:
            # Read first image to see the size
            im = skio.imread(self.filenames[0])
            if self.transform_args is not None:
                im = self.transform_fn(im, *self.transform_args)
            else:
                im = self.transform_fn(im)
            self.img_size = im.shape
        else:
            self.img_size = img_size
        self.arr_size = (len(self.filenames),) + self.img_size
        print("Array size is {}".format(self.arr_size))

    def load(self):
        if self.read_or_create_data(self.arr_size, self.dtype) is not True:
            loader = ParallelImageLoader(self.filenames,
                                         self.data,
                                         transform_fn=self.transform_fn,
                                         transform_args=self.transform_args)
            loader.load()

    def create_generator(self, batch_size, idxs=None, map_fn=TFGenerator.map_fn_divide_255, shuffle=True, one_shot=False, undersample=False):
        # Create generators for training
        gen = TFGenerator(self.data,
                          self.cls,
                          idxs=idxs,
                          batch_size=batch_size,
                          map_fn=map_fn,
                          shuffle=shuffle,
                          one_shot=one_shot,
                          undersample=undersample)
        return gen



import glob
import os
import hashlib
import gc
import numpy as np
import pandas as pd
from PIL import Image
import skimage.color as skcolor
from skimage.transform import resize
from sklearn.model_selection import train_test_split
from tensorflow.keras.utils import to_categorical
from collections import OrderedDict
from scipy.stats.mstats import gmean
from miso.data.download import download_images
from numpy.lib.format import open_memmap
import lxml.etree as ET
from pathlib import Path


class DataSource:

    def __init__(self):
        self.num_classes = 0
        self.cls_labels = []
        self.cls_counts = []
        self.source_name = ""

        self.data_df = None
        self.train_df = None
        self.test_df = None

        self.random_idx = None
        self.random_idx_init = None

        self.images = None
        self.cls = None
        self.onehots = None
        self.vectors = None

        self.train_images = None
        self.train_cls = None
        self.train_onehots = None
        self.train_vectors = None

        self.test_images = None
        self.test_cls = None
        self.test_onehots = None
        self.test_vectors = None

        self.use_mmap = False
        self.images_mmap_filename = None
        self.mmap_directory = None

    def get_class_weights(self):
        count = np.bincount(self.data_df['cls']).astype(np.float64)
        weights = gmean(count) / count
        weights[weights < 0.1] = 0.1
        weights[weights > 10] = 10
        return weights

    def get_short_filenames(self):
        return pd.concat((self.train_df.short_filenames, self.test_df.short_filenames))

    @staticmethod
    def preprocess_image(im, prepro_method='rescale', prepro_params=(255, 0, 1)):
        # TODO divide according to image depth (8 or 16 bits etc)
        if prepro_method == 'rescale':
            im = np.divide(im, prepro_params[0])
            im = np.subtract(im, prepro_params[1])
            im = np.multiply(im, prepro_params[2])
        return im

    @staticmethod
    def load_image(filename, img_size, img_type):
        if img_type == 'rgb':
            im = Image.open(filename)
            im = np.asarray(im, dtype=np.float)
            if im.ndim == 2:
                im = np.expand_dims(im, -1)
                im = np.repeat(im, repeats=3, axis=-1)
        elif img_type == 'greyscale':
            im = Image.open(filename).convert('L')
            im = np.asarray(im, dtype=np.float)
            im = np.expand_dims(im, -1)
        elif img_type == 'greyscale3':
            im = Image.open(filename).convert('L')
            im = np.asarray(im, dtype=np.float)
            im = np.expand_dims(im, -1)
            im = np.repeat(im, repeats=3, axis=-1)
        elif img_type == 'greyscaled':
            ims = DataSource.read_tiff(filename, [0, 2])
            g = skcolor.rgb2grey(ims[0]) * 255      # Scales to 0 - 1 for some reason
            if ims[1].ndim == 3:
                d = skcolor.rgb2grey(ims[1])
            else:
                d = ims[1].astype(float)
            im = np.concatenate((g[:, :, np.newaxis], d[:, :, np.newaxis]), 2)
        elif img_type == 'greyscaledm':
            ims = DataSource.read_tiff(filename, [0, 2, 4])
            g = skcolor.rgb2grey(ims[0]) * 255  # Scales to 0 - 1 for some reason
            if ims[1].ndim == 3:
                d = skcolor.rgb2grey(ims[1])
            else:
                d = ims[1].astype(float)
            if ims[2].ndim == 3:
                m = skcolor.rgb2grey(ims[2])
            else:
                m = ims[2].astype(float)
            im = np.concatenate((g[:, :, np.newaxis], d[:, :, np.newaxis], m[:, :, np.newaxis]), 2)
        elif img_type == 'rgbd':
            ims = DataSource.read_tiff(filename, [0, 2])
            rgb = ims[0]
            # print("rgbd {}".format(filename))
            # print(rgb.shape)
            # print(ims[1].shape)
            if rgb.ndim == 2:
                rgb = np.expand_dims(rgb, -1)
                rgb = np.repeat(rgb, repeats=3, axis=-1)
            d = skcolor.rgb2grey(ims[1])
            im = np.concatenate((rgb, d[:,:, np.newaxis]), 2)
        # print(im.shape)
        im = DataSource.make_image_square(im)
        # print(im.shape)
        im = resize(im, (img_size[0], img_size[1]), order=1)
        # print(im.shape)
        return im

    @staticmethod
    def read_tiff(filename, indices):
        img = Image.open(filename)
        images = []
        # print("num frames: {}".format(img.n_frames))
        for i in range(img.n_frames):
            img.seek(i)
            # print("- frame {} shape {}".format(i, np.array(img).shape))
            # print(img)
        for i, idx in enumerate(indices):
            img.seek(idx)
            if len(np.array(img).shape) == 0:
                #print("Bad")
                img.mode = 'L'
            images.append(np.array(img))
        return images

    def load_dataset(self,
                     img_size,
                     img_type='rgb',
                     dtype=np.float16):

        # Image filenames
        filenames = self.data_df['filenames']
        image_count = len(filenames)

        # Color mode:
        # - rgb: normal RGB
        # - greyscale: convert to greyscale (single channel) if necessary
        # - greyscale3: convert to greyscale then repeat across 3 channels
        #               (for inputting greyscale images into networks that take three channels)
        if img_type == 'rgb' or img_type == 'greyscale3':
            channels = 3
        elif img_type == 'greyscale':
            channels = 1
        elif img_type == 'rgbd':
            channels = 4
        elif img_type == 'greyscaled':
            channels = 2
        elif img_type == 'greyscaledm':
            channels = 3
        else:
            raise ValueError("Unknown image type")

        # float16 is used be default to save memory
        if dtype is np.float16:
            byte_count = 2
        elif dtype is np.float32:
            byte_count = 4
        elif dtype is np.float64:
            byte_count = 8
        else:
            byte_count = 'X'

        # Sometimes the data set is too big to be saved into memory.
        # In this case, we can memory map the numpy array onto disk.
        # Make sure to delete the files afterwards
        if self.use_mmap:
            # Unique hash id is used for the filename
            hashstr = hashlib.sha256(pd.util.hash_pandas_object(self.data_df, index=True).values).hexdigest()[0:16]
            unique_id = "{}_{}_{}_{}_{}.npy".format(hashstr, img_size[0], img_size[1], img_type, byte_count)
            self.images_mmap_filename = os.path.join(self.mmap_directory, unique_id)
            print(self.images_mmap_filename)
            # If the memmap file already exists, simply load it
            if os.path.exists(self.images_mmap_filename):
                self.images = open_memmap(self.images_mmap_filename, dtype=dtype, mode='r+', shape=(image_count, img_size[0], img_size[1], channels))
                return
            self.images = open_memmap(self.images_mmap_filename, dtype=dtype, mode='w+',  shape=(image_count, img_size[0], img_size[1], channels))
        else:
            self.images = np.zeros(shape=(image_count, img_size[0], img_size[1], channels), dtype=dtype)

        # Load each image
        idx = 0
        print("@ Loading images... ")
        for filename in filenames:
            try:
                im = self.load_image(filename, img_size, img_type)
                im = self.preprocess_image(im)
                # Convert to format
                im = im.astype(dtype)
                if im.ndim == 2:
                    im = im[:, :, np.newaxis]
                self.images[idx] = im
            except:
                print("@ Error loading image {}".format(filename))
            idx += 1
            if idx % 100 == 0:
                print("\r@ Loading images {}%".format((int)(idx / len(filenames) * 100)))
        if self.use_mmap:
            self.images.flush()

    def delete_memmap_files(self, del_split=True, del_source=True):
        if self.use_mmap is False:
            return
        if self.mmap_directory is None:
            return
        if del_split:
            train_filename = os.path.join(self.mmap_directory, "train.npy")
            test_filename = os.path.join(self.mmap_directory, "test.npy")
            if os.path.exists(train_filename):
                if self.train_images is not None:
                    self.train_images._mmap.close()
                del self.train_images
                gc.collect()
                os.remove(train_filename)
            if os.path.exists(test_filename):
                if self.test_images is not None:
                    self.test_images._mmap.close()
                del self.test_images
                gc.collect()
                os.remove(test_filename)
        if del_source:
            if os.path.exists(self.images_mmap_filename):
                if self.images is not None:
                    self.images._mmap.close()
                del self.images
                gc.collect()
                os.remove(self.images_mmap_filename)

    def split(self, split=0.20, seed=None):
        dtype=self.images.dtype
        if split > 0.0:
            # Split with stratify
            train_idx, test_idx = train_test_split(range(len(self.images)), test_size=split, random_state=seed, shuffle=True, stratify=self.cls)
            self.random_idx = train_idx + test_idx
        else:
            train_idx = np.random.permutation(range(len(self.images)))
            test_idx = []
            self.random_idx = train_idx
        print("@ Split mapping...")
        img_size = self.images.shape[1:]
        # Memmap splitting
        if self.use_mmap:
            print("@ Split mapping - deleting old memmap files")
            train_filename = os.path.join(self.mmap_directory, "train.npy")
            test_filename = os.path.join(self.mmap_directory, "test.npy")
            self.delete_memmap_files(del_split=True, del_source=False)
            print("@ Split mapping - creating new memmap files")
            self.train_images = open_memmap(train_filename, dtype=dtype, mode='w+', shape=(len(train_idx), ) + img_size)
            self.test_images = open_memmap(test_filename, dtype=dtype, mode='w+', shape=(len(test_idx), ) + img_size)
            print("@ Split mapping - copying train images")
            for i in range(len(train_idx)):
                self.train_images[i] = self.images[train_idx[i]]
            print("@ Split mapping - copying test images")
            for i in range(len(test_idx)):
                self.test_images[i] = self.images[test_idx[i]]
        # Normal splitting
        else:
            self.train_images = self.images[train_idx]
            self.test_images = self.images[test_idx]
        # Remainder
        self.train_cls = self.cls[train_idx]
        self.test_cls = self.cls[test_idx]
        self.train_onehots = self.onehots[train_idx]
        self.test_onehots = self.onehots[test_idx]
        self.train_df = self.data_df.iloc[train_idx,:]
        self.test_df = self.data_df.iloc[test_idx,:]
        print("@ Split mapping - done")

    def set_source(self,
                   source,
                   min_count,
                   max_count=None,
                   min_count_to_others=False,
                   extension=None,
                   mapping: dict = None,
                   map_others=True,
                   must_contain: str = None,
                   ignore_list: list = None,
                   mmap_directory = None):
        """
        Loads images from from a directory where each sub-directories contains images for a single class, e.g.:

        directory
          |-- class 1 directory
          |-- class 2 directory
          |-- class 3 directory
          `-- ...

        The cls for the class are taken as the sub-directory names

        :param source: Path to the directory containing sub-directories of classes
        :param extension: Extension of the images in directory (e.g. "jpg"). If `None`, it looks for jpg, png and tiff
        :param min_count: Minimum number of images in a sub-directory for that class to be included
        :param max_count: Maximum number of images in a sub-directory to be used (If `None` all images are used)
        :param mapping: Dictionary mapping classes to final classes. E.g. {"cat": "animal", "dog",:"animal"} maps "cat" and "dog" both to animal.
        If mapping is `None`, the original classes are used. If mapping is not `None` then only the classes in the map are used.
        :param map_others: If `True` then classes not in the mapping will be mapped to an "Others" class
        :param must_contain: The image filenames must contain this string
        :param ignore_list: List of classes that will be ignored, and their images not loaded
        :return:
        """
        self.mmap_directory = mmap_directory
        if source.startswith("http"):
            print("@ Downloading dataset " + source + "...")
            dir_for_download = os.path.join(os.getcwd(), 'datasets')
            os.makedirs(dir_for_download, exist_ok=True)
            dir_path = download_images(source, dir_for_download)
            self.source_name = dir_path
            if mmap_directory is None:
                self.mmap_directory = dir_for_download
        else:
            self.source_name = source
            if mmap_directory is None:
                self.mmap_directory = str(Path(self.source_name).parent)

        if self.source_name.endswith("xml"):
            print("@ Parsing project file " + self.source_name)
            filenames = self.parse_xml(self.source_name)
            if mmap_directory is None:
                self.mmap_directory = str(Path(self.source_name).parent)
        else:
            print("@ Parsing image directory...")
            # Get alphabetically sorted list of class directories
            class_dirs = sorted(glob.glob(os.path.join(self.source_name, "*")))

            # Load images from each class and place into a dictionary
            filenames = OrderedDict()
            for class_dir in class_dirs:
                if os.path.isdir(class_dir) is False:
                    continue

                # Get the class name
                class_name = os.path.basename(class_dir)

                # Skip directories starting with ~
                if class_name.startswith('~'):
                    continue

                # Get the files
                files = []
                if extension is None:
                    for ext in ["*.jpg", "*.jpeg", "*.png", "*.tif", "*.tiff", "*.bmp"]:
                        if must_contain is not None:
                            files.extend(sorted(glob.glob(os.path.join(class_dir, "*" + must_contain + ext))))
                        else:
                            files.extend(sorted(glob.glob(os.path.join(class_dir, ext))))
                else:
                    if must_contain is not None:
                        files = sorted(glob.glob(os.path.join(class_dir, "*" + must_contain + "*." + extension)))
                    else:
                        files = sorted(glob.glob(os.path.join(class_dir, "*." + extension)))

                # Add to dictionary
                filenames[class_name] = files

        # Map the classes into overall classes if mapping is enabled
        if mapping is not None:
            print("@ Applying mapping...")
            mapped_filenames = OrderedDict()
            # Sort the map
            sorted_map = OrderedDict()
            for key in sorted(mapping.keys()):
                sorted_map[key] = mapping[key]

            # Keep a record of which classes have already been mapped
            already_mapped_cls = list()

            # Iterate through the map
            for cls, sub_cls_list in sorted_map.items():
                print(" - {} <= ".format(cls), end='')
                # Create list entry for this class
                mapped_filenames[cls] = list()
                # Add all the component classes
                for sub_cls in sub_cls_list:
                    if sub_cls in filenames and sub_cls not in already_mapped_cls:
                        print("{} ".format(sub_cls), end='')
                        mapped_filenames[cls].extend(filenames[sub_cls])
                        already_mapped_cls.append(sub_cls)
                # Number of classes
                print("({} images)".format(len(mapped_filenames[cls])))

            # Add others
            if map_others is True:
                mapped_filenames['other'] = list()
                # Iterate though the filenames dictionary and add any classes not already mapped to others
                for cls in filenames.keys():
                    if cls not in already_mapped_cls:
                        mapped_filenames['other'].extend(filenames[cls])

            # Save the mapped filenames as the current filenames dictionary
            filenames = mapped_filenames

        # Remove any classes that do not have enough images and put in 'other'
        print("@ Moving classes with not enough images to 'other'...")
        not_enough_list = list()
        enough_dict = OrderedDict()
        for cls, cls_filenames in filenames.items():
            print(" - ({:5d} images) {}".format(len(cls_filenames), cls), end='')
            if len(cls_filenames) < min_count:
                not_enough_list.extend(cls_filenames)
                print(" => other".format(len(cls_filenames)))
            else:
                enough_dict[cls] = cls_filenames
                print()

        # Put the others in the list if there is also enough for them
        if min_count_to_others is True:
            if len(not_enough_list) > min_count:
                if 'other' in enough_dict:
                    enough_dict['other'].extend(not_enough_list)
                else:
                    enough_dict['other'] = not_enough_list
                print(" - {} images in other".format(len(not_enough_list)))
            else:
                print(" - other not included ({} images)".format(len(not_enough_list)))

        filenames = enough_dict
        # print(enough_dict.keys())

        # Finally, create a list for each (make sure 'other' is last class
        cls_index = []
        cls_labels = []
        cls_counts = []
        long_filenames = []
        short_filenames = []
        self.cls_labels = []

        if 'other' in filenames:
            other_index = len(filenames) - 1
        else:
            other_index = len(filenames)

        index = 0
        for cls, cls_filenames in filenames.items():
            for filename in cls_filenames:
                if cls != 'other':
                    cls_index.append(index)
                else:
                    cls_index.append(other_index)
                cls_labels.append(cls)
                long_filenames.append(filename)
                short_filenames.append(os.path.basename(filename))

            if cls != 'other':
                self.cls_labels.append(cls)
                index += 1

        if 'other' in filenames:
            self.cls_labels.append('other')
        df = {"filenames": long_filenames, "short_filenames": short_filenames, "cls": cls_labels, "cls": cls_index}

        self.data_df = pd.DataFrame(df)
        self.num_classes = len(self.cls_labels)
        print("@ {} images in {} classes".format(len(cls_index), self.num_classes))

        for idx in range(self.num_classes):
            cls_counts.append(len(self.data_df[self.data_df['cls'] == idx]))
            # print(cls_counts)
        self.cls_counts = cls_counts
        self.cls = self.data_df['cls'].to_numpy()
        self.onehots = to_categorical(self.data_df['cls'])

        # print(self.data_df)

    @staticmethod
    def make_image_square(im):
        if im.shape[0] == im.shape[1]:
            return im
        height = im.shape[0]
        width = im.shape[1]
        half = max(height, width)
        height_pad_start = int(abs(np.floor((height - half) / 2)))
        height_pad_end = int(abs(np.ceil((height - half) / 2)))
        width_pad_start = int(abs(np.floor((width - half) / 2)))
        width_pad_end = int(abs(np.ceil((width - half) / 2)))
        consts = [np.median(np.concatenate((im[0, :, i], im[-1, :, i], im[:, 0, i], im[:, -1, i]))) for i in
                  range(im.shape[2])]
        im = np.stack(
            [np.pad(im[:, :, c],
                    ((height_pad_start, height_pad_end), (width_pad_start, width_pad_end)),
                    mode='constant',
                    constant_values=consts[c])
             for c in range(im.shape[2])], axis=2)
        return im

    @staticmethod
    def parse_xml(xml_filename):
        project = ET.parse(xml_filename).getroot()

        filenames = []
        cls = []
        cls_labels = []

        filenames_dict = OrderedDict()

        images_xml = project.find('images')
        for i, image_xml in enumerate(images_xml.iter('image')):
            relfile = image_xml.find('source').find('filename').text
            if os.path.isabs(relfile):
                absfile = relfile
            else:
                absfile = os.path.abspath(os.path.join(os.path.dirname(xml_filename), relfile))
            if os.path.isfile(absfile) is False:
                continue
            filenames.append(absfile)

            cls_names = []
            cls_scores = []
            cls_base = image_xml.find('classifications')
            for cls_val in cls_base.iter('classification'):
                cls_names.append(cls_val.find('code').text)
                cls_scores.append(float(cls_val.find('value').text))
            cls.append(cls_names[np.argmax(cls_scores)])

        for taxon_xml in project.find('taxons').iter('taxon'):
            if taxon_xml.find('isClass').text == 'true':
                cls_labels.append(taxon_xml.find('code').text)
        cls_labels = sorted(cls_labels)
        
        df = pd.DataFrame({'filenames': filenames, 'cls': cls})
        for label in cls_labels:
            filenames_dict[label] = df.filenames[df.cls == label]
        return filenames_dict

    @staticmethod
    def parse_directory(source_dir):
        sub_dirs = sorted(glob.glob(os.path.join(source_dir, "*")))

        filenames = OrderedDict()
        for sub_dir in sub_dirs:
            if os.path.isdir(sub_dir) is False:
                continue
            # Get the directory name
            sub_name = os.path.basename(sub_dir)
            # Skip directories starting with ~
            if sub_name.startswith('~'):
                continue
            # Get the files
            files = []
            for ext in ["*.jpg", "*.jpeg", "*.png", "*.tif", "*.tiff", "*.bmp"]:
                files.extend(sorted(glob.glob(os.path.join(sub_dir, ext))))
            # Add to dictionary
            filenames[sub_name] = files
        return filenames


if __name__ == "__main__":
    ds = DataSource()
    ds.set_source(r"C:\Users\rossm\Documents\Data\Foraminifera\BenthicPlanktic\Benthic_Planktic_Source_v2\project.xml", 40)



from sklearn import discriminant_analysis, manifold
from sklearn.decomposition import PCA
from sklearn.preprocessing import normalize
import numpy as np


def pca(X, nr_components=16, normalise_vectors=True):
    if normalise_vectors:
        X = normalize(X, axis=1)
    p = PCA(n_components=nr_components)
    p.fit(X)
    print(p.explained_variance_ratio_)
    print(np.sum(p.explained_variance_ratio_))
    return p.transform(X)


def lda(X, y, nr_components=2):
    """
    Linear discrimindant analysis
    :param X: Input vectors
    :param y: Input classes
    :param nr_components: Dimension of output co-ordinates
    :return: Output co-ordinates
    """
    print("Computing Linear Discriminant Analysis projection")
    X2 = X.copy()
    X2.flat[::X.shape[1] + 1] += 0.01  # Make X invertible
    return discriminant_analysis.LinearDiscriminantAnalysis(n_components=nr_components).fit_transform(X2, y)


def tsne(X, nr_components=2, perplexity=30, early_exaggeration=12):
    """
    t-SNE clustering
    :param X: Input vector
    :param nr_components: Dimension of output co-ordinates
    :param perplexity: t-SNE perplexity
    :param early_exaggeration: t-SNE early exaggeration
    :return: Output co-ordinates
    """
    print("Computing t-SNE embedding")
    tsne = manifold.TSNE(n_components=nr_components, init='random', random_state=0, perplexity=perplexity, early_exaggeration=early_exaggeration)
    return tsne.fit_transform(X)



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='miso',
    version='3.1.20',
    description='Python scripts for training CNNs for particle classification',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ross Marchant',
    author_email='ross.g.marchant@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
    ],
    keywords='microfossil, cnn',
    python_requires='>=3.9,<=3.11',
    packages=find_packages(),
    install_requires=['tensorflow==2.10.1',
                      'image-classifiers==1.0.0',
                      'lxml==5.1.0',
                      'matplotlib==3.8.2',
                      'numpy==1.26.4',
                      'pandas==2.1.4',
                      'Pillow==10.2.0',
                      'imagecodecs==2024.1.1',
                      'scikit-image==0.22.0',
                      'scikit-learn==1.3.2',
                      'scipy==1.11.4',
                      'segmentation-models==1.0.1',
                      'dill==0.3.7',
                      'flask==3.0.0',
                      'itsdangerous==2.1.2',
                      'tqdm==4.66.1',
                      'openpyxl==3.1.2',
                      'imbalanced-learn==0.11.0',
                      'onnx==1.12.0',
                      'tf2onnx==1.14.0',
                      'protobuf==3.19.6',
                      'cleanlab==2.5.0',
                      'packaging==23.2',
                      'tensorflow_addons==0.19.0',
                      'marshmallow_dataclass==8.6.0',
                      'opencv-python==4.9.0.80',
                      'onnxruntime-gpu==1.17.0'],
    url='https://github.com/microfossil/particle-classification',
    license='MIT',
    project_urls={  # Optional
        'Source': 'https://github.com/microfossil/particle-classification',
        'Paper': 'https://jm.copernicus.org/articles/39/183/2020/',
    },
)



import collections
import tensorflow as tf

from keras import applications as ka, Input, Model
from keras.layers import Lambda, Dropout, Dense, BatchNormalization

from miso.layers.cyclic import CyclicGainSlice12, CyclicDensePoolN, CyclicSlice4, CyclicDensePool4

KerasModelParameters = collections.namedtuple(
    'KerasModelParameters',
    ['model_func', 'prepro_func', 'default_input_shape']
)


def tf_prepro(input_shape):
    # (x / 127.5) - 1
    inputs = Input(shape=input_shape)
    x = Lambda(lambda y: tf.subtract(tf.multiply(y, 2), 1))(inputs)
    return inputs, x


def torch_prepro(input_shape):
    # (x / 255 - mean) / std_dev
    inputs = Input(shape=input_shape)
    x = Lambda(lambda y: tf.subtract(y, tf.reshape(tf.constant([0.485, 0.456, 0.406]), [1, 1, 1, 3])))(inputs)
    x = Lambda(lambda y: tf.divide(y, tf.reshape(tf.constant([0.229, 0.224, 0.225]), [1, 1, 1, 3])))(x)
    return inputs, x


def caffe_prepro(input_shape):
    # Convert to BGR then x - mean
    inputs = Input(shape=input_shape)
    x = Lambda(lambda y: tf.reverse(y, axis=[-1]))(inputs)
    x = Lambda(lambda y: tf.subtract(tf.multiply(y, 255.0), tf.reshape(tf.constant([103.939, 116.779, 128.68]), [1, 1, 1, 3])))(x)
    return inputs, x


def no_prepro(input_shape):
    inputs = Input(shape=input_shape)
    x = Lambda(lambda y: tf.multiply(y, 255.0))(inputs)
    return inputs, x


def head(cnn_type, use_cyclic, use_gain, input_shape, weights='imagenet', add_prepro=False, trainable=True):
    if cnn_type not in KERAS_MODEL_PARAMETERS:
        raise ValueError(f"CNN type {cnn_type} is not supported, valid values are {KERAS_MODEL_PARAMETERS.keys()}")

    params = KERAS_MODEL_PARAMETERS[cnn_type]
    if add_prepro:
        inputs, x = params.prepro_func(input_shape)
    else:
        inputs = Input(shape=input_shape)
        x = inputs
    if use_cyclic:
        if use_gain:
            x = CyclicGainSlice12()(x)
        else:
            x = CyclicSlice4()(x)
    x = params.model_func(include_top=False, weights=weights, pooling='avg', input_shape=input_shape).call(x)
    if use_cyclic:
        if use_gain:
            x = CyclicDensePoolN(pool_op=tf.reduce_mean)(x)
        else:
            x = CyclicDensePool4(pool_op=tf.reduce_mean)(x)
    model = Model(inputs=inputs, outputs=x)
    if not trainable:
        for layer in model.layers:
            layer.trainable = False
    return model


def tail(num_classes, input_shape, dropout=(0.4, 0.3)):
    # inp = Input(shape=input_shape)
    # x = Dropout(rate=dropout[0], name='dropout1')(inp)
    # x = BatchNormalization()(x)
    # x = Dense(512, activation='relu', bias_initializer='zeros')(x)
    # x = Dropout(rate=dropout[1], name='dropout2')(x)
    # x = BatchNormalization()(x)
    # outp = Dense(num_classes, activation='softmax', kernel_initializer='random_uniform', bias_initializer='zeros')(x)

    # inp = Input(shape=input_shape)
    # x = Dropout(rate=0.5)(inp)
    # # x = BatchNormalization()(x)
    # x = Dense(512, activation='relu', bias_initializer='zeros')(x)
    # # x = Dropout(rate=dropout[1], name='dropout2')(x)
    # # x = BatchNormalization()(x)
    # outp = Dense(num_classes, activation='softmax', kernel_initializer='random_uniform', bias_initializer='zeros')(x)

    inp = Input(shape=input_shape)
    outp = Dropout(dropout[0])(inp)
    outp = Dense(512, activation='relu')(outp)
    # outp = Dropout(dropout[1])(outp)
    # outp = Dense(512, activation='relu')(outp)
    outp = Dense(num_classes, activation='softmax')(outp)

    # outp = Dense(num_classes, activation='softmax')(inp)

    return Model(inp, outp)


def tail_vector(num_classes, input_shape, dropout=(0.5, 0.5)):
    inp = Input(shape=input_shape)
    outp = Dropout(dropout[0])(inp)
    outp = Dense(512, activation='relu')(outp)
    # outp = Dropout(dropout[1])(outp)
    # outp = Dense(512, activation='relu')(outp)
    return Model(inp, outp)


KERAS_MODEL_PARAMETERS = {
    'xception': KerasModelParameters(ka.xception.Xception, tf_prepro, [299, 299, 3]),
    'vgg16': KerasModelParameters(ka.vgg16.VGG16, caffe_prepro, [224, 224, 3]),
    'vgg19': KerasModelParameters(ka.vgg19.VGG19, caffe_prepro, [224, 224, 3]),
    'resnet50': KerasModelParameters(ka.resnet.ResNet50, caffe_prepro, [224, 224, 3]),
    'resnet50v2': KerasModelParameters(ka.resnet_v2.ResNet50V2, tf_prepro, [224, 224, 3]),
    'resnet101': KerasModelParameters(ka.resnet.ResNet101, caffe_prepro, [224, 224, 3]),
    'resnet101v2': KerasModelParameters(ka.resnet_v2.ResNet101V2, tf_prepro, [224, 224, 3]),
    'resnet152': KerasModelParameters(ka.resnet.ResNet152, caffe_prepro, [224, 224, 3]),
    'resnet152v2': KerasModelParameters(ka.resnet_v2.ResNet152V2, tf_prepro, [224, 224, 3]),
    'inceptionresnetV2': KerasModelParameters(ka.inception_resnet_v2.InceptionResNetV2, tf_prepro, [299, 299, 3]),
    'mobilenet': KerasModelParameters(ka.mobilenet.MobileNet, tf_prepro, [224, 224, 3]),
    'mobilenetV2': KerasModelParameters(ka.mobilenet_v2.MobileNetV2, tf_prepro, [224, 224, 3]),
    'densenet121': KerasModelParameters(ka.densenet.DenseNet121, torch_prepro, [224, 224, 3]),
    'densenet169': KerasModelParameters(ka.densenet.DenseNet169, torch_prepro, [224, 224, 3]),
    'densenet201': KerasModelParameters(ka.densenet.DenseNet201, torch_prepro, [224, 224, 3]),
    'nasnetmobile': KerasModelParameters(ka.nasnet.NASNetMobile, tf_prepro, [224, 224, 3]),
    'nasnetlarge': KerasModelParameters(ka.nasnet.NASNetLarge, tf_prepro, [331, 331, 3]),
    'efficientnetb0': KerasModelParameters(ka.efficientnet.EfficientNetB0, no_prepro, [224, 224, 3]),
    'efficientnetb1': KerasModelParameters(ka.efficientnet.EfficientNetB1, no_prepro, [240, 240, 3]),
    'efficientnetb2': KerasModelParameters(ka.efficientnet.EfficientNetB2, no_prepro, [260, 260, 3]),
    'efficientnetb3': KerasModelParameters(ka.efficientnet.EfficientNetB3, no_prepro, [300, 300, 3]),
    'efficientnetb4': KerasModelParameters(ka.efficientnet.EfficientNetB4, no_prepro, [380, 380, 3]),
    'efficientnetb5': KerasModelParameters(ka.efficientnet.EfficientNetB5, no_prepro, [456, 456, 3]),
    'efficientnetb6': KerasModelParameters(ka.efficientnet.EfficientNetB6, no_prepro, [528, 528, 3]),
    'efficientnetb7': KerasModelParameters(ka.efficientnet.EfficientNetB7, no_prepro, [600, 600, 3]),
    'convnexttiny': KerasModelParameters(ka.convnext.ConvNeXtTiny, no_prepro, [224, 224, 3]),
    'convnextsmall': KerasModelParameters(ka.convnext.ConvNeXtSmall, no_prepro, [224, 224, 3]),
    'convnextbase': KerasModelParameters(ka.convnext.ConvNeXtBase, no_prepro, [224, 224, 3]),
    'convnextlarge': KerasModelParameters(ka.convnext.ConvNeXtLarge, no_prepro, [224, 224, 3]),
    'convnextxlarge': KerasModelParameters(ka.convnext.ConvNeXtXLarge, no_prepro, [224, 224, 3]),
}



import tensorflow as tf
from tensorflow.keras.callbacks import Callback
import tensorflow.keras.backend as K
from miso.utils.rolling_buffer import RollingBuffer
import math
import time


def graph_to_console(epoch, batch, acc, loss, val_acc, val_loss, lr_prob, lr_prob_active, time_difference):
    acc_i = round(acc * 50)
    val_acc_i = round(val_acc * 50)
    lr_prob_i = round(lr_prob * 50)

    for j in range(51):
        if j == acc_i:
            print('#', end="")
        # elif j == trainsdi:
        #     print('@', end="")
        elif j == val_acc_i:
            print('*', end="")
        elif j == lr_prob_i and j != 100 and lr_prob_active:
            print('+', end="")
        elif j % 10 == 0:
            print('|', end="")
        else:
            print(' ', end="")
    # msg = " {} #T {:.1f}%/{:.4f}, *V {:.1f}%/{:.4f} ({:.2f}s)"
    # print(msg.format(epoch, acc * 100, loss, val_acc * 100, val_loss, time_difference))
    msg = " {} #T{:.1f}%/*V{:.1f}% ({:.2f}s)"
    print(msg.format(epoch, acc * 100, val_acc * 100, time_difference))


class AdaptiveLearningRateScheduler(Callback):
    """
    Adaptive learning rate scheduler

    Decreases learning rate by a certain factor each time it is no longer improving
    """

    def __init__(self, drop_rate=0.5, nb_drops=4, nb_epochs=10, verbose=1, monitor='loss'):
        super(AdaptiveLearningRateScheduler, self).__init__()
        self.monitor = monitor
        self.drop_rate = drop_rate
        self.nb_drops = nb_drops
        self.nb_epochs = nb_epochs
        self.verbose = verbose
        self.current_epoch = 0
        self.current_batch = 0
        self.drop_count = 0
        self.buffer = None
        self.previous_time = None
        self.finished = False

    def on_train_begin(self, logs=None):
        # if 'batch_size' in self.params and self.params['batch_size'] is not None:
        #     batch_size = self.params['batch_size']
        #     samples = self.params['samples']
        #     self.buffer = RollingBuffer(math.ceil(samples * self.nb_epochs / batch_size))
        # else:
        self.buffer = RollingBuffer(self.nb_epochs)
        self.previous_time = time.time()

    def on_epoch_begin(self, epoch, logs=None):
        self.current_epoch = epoch

    def on_epoch_end(self, epoch, logs=None):
        loss = logs.get("loss")
        acc = logs.get("acc") or logs.get('accuracy') or logs.get('iou_score') or logs.get('cosine_proximity')
        val_loss = logs.get("val_loss") or 0
        val_acc = logs.get("val_acc") or logs.get('val_accuracy') or logs.get('val_iou_score') or logs.get(
            'val_cosine_proximity') or 0
        if 'cosine_proximity' in logs:
            acc += 1
            val_acc += 1

        # Elapsed time
        current_time = time.time()
        if self.previous_time is None:
            time_difference = 0
        else:
            time_difference = current_time - self.previous_time
        self.previous_time = current_time

        # Update learning rate
        self.update_learning_rate(self.current_epoch, logs)

        # Graph to console
        if val_acc is not None:
            print("\r", end="")
            graph_to_console(self.current_epoch, self.current_batch,
                             acc, loss, val_acc, val_loss,
                             self.buffer.slope_probability_less_than(0), self.buffer.full(),
                             time_difference)

        if self.finished is True:
            self.model.stop_training = True
            print("Training finished".format(self.model.optimizer.lr))

    def on_batch_end(self, batch, logs=None):
        self.current_batch += 1

    def update_learning_rate(self, count, logs):
        monitor_value = logs.get(self.monitor)
        self.buffer.append(monitor_value)

        if count >= self.buffer.length() * 3 and self.buffer.full() and self.finished is False:
            # if count % 20 == 19:
            #     lr = float(K.get_value(self.model.optimizer.lr))
            #     new_lr = lr * self.drop_rate
            #     K.set_value(self.model.optimizer.lr, new_lr)
            #     print("Learning rate dropped ({}/{}) to {}".format(self.drop_count, self.nb_drops, new_lr))
            if self.buffer.slope_probability_less_than(0) < 0.50:
                lr = float(K.get_value(self.model.optimizer.lr))
                # lr = self.model.optimizer.lr.read_value()
                new_lr = lr * self.drop_rate
                K.set_value(self.model.optimizer.lr, new_lr)
                # self.model.optimizer.lr.assign(new_lr)
                self.buffer.clear()
                self.drop_count += 1
                if self.drop_count == self.nb_drops:
                    self.finished = True
                    return
                if self.verbose == 1:
                    print("Learning rate dropped ({}/{}) to {}".format(self.drop_count, self.nb_drops, new_lr))



import click

# Cant have name same as cli method name
from miso.inference.classify import classify_folder as miso_classify_folder, segment_folder
from miso.inference.classify import segment_folder as miso_segment_folder


@click.group()
def cli():
    pass


@cli.command()
@click.option('--model', '-m', type=click.Path(exists=True), required=True, help='Path to the model information.')
@click.option('--input', '-i', type=click.Path(exists=True), required=True, help='Path to the directory containing images.')
@click.option('--output', '-o', type=click.Path(), required=True, help='Path where the output CSV will be saved.')
@click.option('--batch_size', '-b', type=int, default=32, show_default=True, help='Batch size for processing images.')
@click.option('--sample', '-s', type=str, default='unknown', show_default=True, help='Default sample name')
@click.option('--unsure_threshold', '-u', type=float, default=0.0, show_default=True, help='Threshold below which predictions are considered unsure.')
def classify_folder(model, input, output, batch_size, sample, unsure_threshold):
    """
    Classify images in a folder and output the results to a CSV file.
    """
    miso_classify_folder(model, input, output, batch_size, sample, unsure_threshold)


@cli.command()
@click.option('--model', '-m', type=click.Path(exists=True), required=True, help='Path to the model information.')
@click.option('--input', '-i', type=click.Path(exists=True), required=True, help='Path to the folder containing images.')
@click.option('--output', '-o', type=click.Path(), required=True, help='Path where the morphology csv will be saved.')
@click.option('--batch_size', '-b', type=int, default=32, show_default=True, help='Batch size for processing images.')
@click.option('--sample', '-s', type=str, default='unknown', show_default=True, help='Default sample name')
@click.option('--threshold', '-t', type=float, default=0.5, show_default=True, help='Threshold for segmentation.')
@click.option('--save-contours', is_flag=True, default=False, help='Whether to save contours or not.')
def segment_folder(model, input, output, batch_size, sample, threshold, save_contours):
    """
    Segment images in a folder and output the results.
    """
    miso_segment_folder(model, input, output, batch_size, sample, threshold, save_contours)



if __name__ == "__main__":
    cli()


# if __name__ == "__main__":
#     parser = argparse.ArgumentParser(description="Train a CNN to classify images")
#     parser.add_argument("-i", "--input", required=True, help="Directory of images, URL link to zipped directory of images, or ParticleTrieur project file")
#     parser.add_argument("-o", "--output", required=True, help="Output directory to store training results")
#     parser.add_argument("-t", "--type", required=True, help="Type of CNN: resnet_tl, resnet_cyclic_tl, base_cyclic, resnet_cyclic, resnet18, resnet50, vgg16, vgg19")
#
#     parser.add_argument("-f", "--filters", type=int, default=4, help="Number of filters in the first convolutional block")
#
#     parser.add_argument("--min_count", type=int, default=10, help="Minimum number of images in a class for it to be included")
#     parser.add_argument("--map_others", action='store_true', help="Classes with not enough images will be put into 'others' class (so long as the total is also greater than min_count")
#     args = parser.parse_args()
#
#     tp = MisoParameters()
#     tp.source = args.input
#     tp.save_dir = args.output
#     tp.cnn_type = args.type
#     tp.filters = args.filters
#     tp.min_count = args.min_count
#     tp.map_others = args.map_others
#
#     train_image_classification_model(tp)
    # print("This is the name of the script: ", sys.argv[0])
    # print("Number of arguments: ", len(sys.argv))
    # print("The arguments are: ", str(sys.argv))
    #
    # params = default_params()
    # print(params)
    # key = None
    # for arg in sys.argv[1:]:
    #     print(arg)
    #     if arg.startswith('--'):
    #         key = arg[2:]
    #         print(key)
    #     else:
    #         try:
    #             val = int(arg)
    #             print("int")
    #         except ValueError:
    #             try:
    #                 val = float(arg)
    #                 print("float")
    #             except:
    #                 val = arg
    #                 print("string")
    #         params[key] = val
    # print(params)
    # train_image_classification_model(params)




from multiprocessing import Process, cpu_count, Queue
import skimage.io as skio
import numpy as np
from tqdm import tqdm


def produce(producer_queue, data, workers):
    for datum in data:
        producer_queue.put(datum)
    [producer_queue.put(None) for i in range(workers)]


def work(producer_queue, consumer_queue, transform_fn, transform_args):
    while True:
        res = producer_queue.get()
        if res is None:
            consumer_queue.put(None)
            break
        im = skio.imread(res[1])
        if transform_fn is not None:
            if transform_args is not None:
                im = transform_fn(im, *transform_args)
            else:
                im = transform_fn(im)
        consumer_queue.put((res[0], im))


def consume(consumer_queue, array, multiplier, num_workers):
    pbar = tqdm(total=len(array))
    while num_workers > 0:
        res = consumer_queue.get()
        if res is None:
            num_workers -= 1
            continue
        array[res[0]*multiplier:(res[0]+1)*multiplier] = res[1]
        pbar.update(multiplier)
    pbar.close()


def get_array_shape(filenames):
    im = skio.imread(filenames[0])
    return (len(filenames),) + im.shape


class ParallelImageLoader:
    def __init__(self, filenames, array, multiplier=1, transform_fn=None, transform_args=None):
        self.filenames = filenames
        self.array = array
        self.transform_fn = transform_fn
        self.transform_args = transform_args
        self.multiplier = multiplier

        self.producer_queue = Queue()
        self.consumer_queue = Queue(cpu_count() * 4)
        self.workers = None
        self.NUMBER_OF_PROCESSES = cpu_count()

    def load(self):
        print("Starting queue with {} workers".format(self.NUMBER_OF_PROCESSES))
        payload = zip(range(len(self.filenames)), self.filenames)
        self.workers = [
            Process(target=work, args=(self.producer_queue, self.consumer_queue, self.transform_fn, self.transform_args), name='producer {}'.format(i)) for i in
            range(self.NUMBER_OF_PROCESSES)]
        for w in self.workers:
            w.start()
        try:
            produce(self.producer_queue, payload, self.NUMBER_OF_PROCESSES)
            consume(self.consumer_queue, self.array, self.multiplier, self.NUMBER_OF_PROCESSES)
        except KeyboardInterrupt:
            print("Keyboard interrupt...")
        finally:
            for w in self.workers:
                w.terminate()
                w.join()
        self.producer_queue.close()
        self.consumer_queue.close()




from tensorflow.keras.layers import Input, Flatten, Dense, Dropout, Lambda
from tensorflow.keras.models import Model
import tensorflow.keras.backend as K
from tensorflow.keras.optimizers import RMSprop
import numpy as np
import random
"""
Modified from https://keras.io/examples/mnist_siamese/
"""


def vector_network(input_shape):
    input = Input(shape=input_shape)
    x = Flatten()(input)
    x = Dense(128, activation='relu')(x)
    x = Dropout(0.1)(x)
    x = Dense(128, activation='relu')(x)
    x = Dropout(0.1)(x)
    x = Dense(128, activation='relu')(x)
    return Model(input, x)


def euclidean_distance(vects):
    x, y = vects
    sum_square = K.sum(K.square(x - y), axis=1, keepdims=True)
    return K.sqrt(K.maximum(sum_square, K.epsilon()))


def eucl_dist_output_shape(shapes):
    shape1, shape2 = shapes
    return (shape1[0], 1)


def contrastive_loss(y_true, y_pred):
    '''Contrastive loss from Hadsell-et-al.'06
    http://yann.lecun.com/exdb/publis/pdf/hadsell-chopra-lecun-06.pdf
    '''
    margin = 1
    square_pred = K.square(y_pred)
    margin_square = K.square(K.maximum(margin - y_pred, 0))
    return K.mean(y_true * square_pred + (1 - y_true) * margin_square)


def compute_accuracy(y_true, y_pred):
    '''Compute classification accuracy with a fixed threshold on distances.
    '''
    pred = y_pred.ravel() < 0.5
    return np.mean(pred == y_true)


def accuracy(y_true, y_pred):
    '''Compute classification accuracy with a fixed threshold on distances.
    '''
    return K.mean(K.equal(y_true, K.cast(y_pred < 0.5, y_true.dtype)))


def create_pairs(x, digit_indices):
    '''Positive and negative pair creation.
    Alternates between positive and negative pairs.
    '''
    pairs = []
    labels = []
    n = min([len(digit_indices[d]) for d in range(num_classes)]) - 1
    for d in range(num_classes):
        print(d)
        for i in range(n):
            print(i)
            z1, z2 = digit_indices[d][i], digit_indices[d][i + 1]
            pairs += [[x[z1], x[z2]]]
            inc = random.randrange(1, num_classes)
            dn = (d + inc) % num_classes
            z1, z2 = digit_indices[d][i], digit_indices[dn][i]
            pairs += [[x[z1], x[z2]]]
            labels += [1, 0]
    return np.array(pairs), np.array(labels)


if __name__ == "__main__":
    from miso.archive.datasource import DataSource
    from miso.models.transfer_learning import head

    # num_classes = 10
    epochs = 200

    # The data, split between train and test sets
    ds = DataSource()
    ds.set_source(r'D:\Datasets\Foraminifera\images_20200226_114546 Fifth with background', 40)
    ds.use_mmap = True
    ds.mmap_directory = r'D:\Temp'
    ds.load_dataset((224,224), img_type='greyscale3', dtype=np.float32)
    ds.split(0.2)

    tl = head('resnet50', (224, 224, 3))
    temp = []
    for i in np.arange(0, len(ds.train_images), 1000):
        print(i)
        temp.append(tl.predict(ds.train_images[i:i+1000]))
    ds.train_vectors = np.concatenate(temp, axis=0)
    temp = []
    for i in np.arange(0, len(ds.test_images), 1000):
        print(i)
        temp.append(tl.predict(ds.test_images[i:i + 1000]))
    ds.test_vectors = np.concatenate(temp, axis=0)

    # (x_train, y_train) = (np.reshape(ds.train_images, (len(ds.train_images), -1)), ds.train_cls)
    # (x_test, y_test) = (np.reshape(ds.test_images, (len(ds.test_images), -1)), ds.test_cls)
    (x_train, y_train) = (ds.train_vectors, ds.train_cls)
    (x_test, y_test) = (ds.test_vectors, ds.test_cls)
    x_train = x_train.astype('float32')
    x_test = x_test.astype('float32')
    x_train /= 255
    x_test /= 255
    input_shape = x_train.shape[1:]
    num_classes = ds.num_classes

    # create training+test positive and negative pairs
    digit_indices = [np.where(y_train == i)[0] for i in range(num_classes)]
    tr_pairs, tr_y = create_pairs(x_train, digit_indices)

    digit_indices = [np.where(y_test == i)[0] for i in range(num_classes)]
    te_pairs, te_y = create_pairs(x_test, digit_indices)

    # network definition
    base_network = vector_network(input_shape)

    input_a = Input(shape=input_shape)
    input_b = Input(shape=input_shape)

    # because we re-use the same instance `base_network`,
    # the weights of the network
    # will be shared across the two branches
    processed_a = base_network(input_a)
    processed_b = base_network(input_b)

    distance = Lambda(euclidean_distance,
                      output_shape=eucl_dist_output_shape)([processed_a, processed_b])

    model = Model([input_a, input_b], distance)

    # train
    rms = RMSprop()
    model.compile(loss=contrastive_loss, optimizer=rms, metrics=[accuracy])
    model.fit([tr_pairs[:, 0], tr_pairs[:, 1]], tr_y,
              batch_size=128,
              epochs=epochs,
              validation_data=([te_pairs[:, 0], te_pairs[:, 1]], te_y))

    # compute final accuracy on training and test sets
    y_pred = model.predict([tr_pairs[:, 0], tr_pairs[:, 1]])
    tr_acc = compute_accuracy(tr_y, y_pred)
    y_pred = model.predict([te_pairs[:, 0], te_pairs[:, 1]])
    te_acc = compute_accuracy(te_y, y_pred)

    print('* Accuracy on training set: %0.2f%%' % (100 * tr_acc))
    print('* Accuracy on test set: %0.2f%%' % (100 * te_acc))

    from sklearn.neighbors import KNeighborsClassifier
    neigh = KNeighborsClassifier(n_neighbors=5)
    neigh.fit(base_network.predict(ds.train_vectors), ds.train_cls)
    yc_pred = neigh.predict(base_network.predict(ds.test_vectors))
    from sklearn.metrics import accuracy_score
    print(accuracy_score(ds.test_cls, yc_pred))

    neigh = KNeighborsClassifier(n_neighbors=5)
    neigh.fit(ds.train_vectors, ds.train_cls)
    yc_pred = neigh.predict(ds.test_vectors)
    from sklearn.metrics import accuracy_score
    print(accuracy_score(ds.test_cls, yc_pred))